import functools

import ipywidgets as widgets


# This will make a little control panel to help us align things
def make_control_panel(m):
    def create_expanded_button(description):
        return widgets.Button(
            description=description,
            button_style="warning",
            layout=widgets.Layout(height="auto", width="auto"),
        )

    def button_move(b, dx=0, dy=0):
        m.move(dx=dx, dy=dy)

    control_panel = widgets.GridspecLayout(7, 7)

    # y buttons
    bp5y = create_expanded_button("+5mm")
    bp5y.on_click(functools.partial(button_move, dx=0, dy=5))
    control_panel[0, 3] = bp5y

    bp1y = create_expanded_button("+1mm")
    bp1y.on_click(functools.partial(button_move, dx=0, dy=1))
    control_panel[1, 3] = bp1y

    bppt1y = create_expanded_button("+.1mm")
    bppt1y.on_click(functools.partial(button_move, dx=0, dy=0.1))
    control_panel[2, 3] = bppt1y

    bmpt1y = create_expanded_button("-.1mm")
    bmpt1y.on_click(functools.partial(button_move, dx=0, dy=-0.1))
    control_panel[4, 3] = bmpt1y

    bm1y = create_expanded_button("-1mm")
    bm1y.on_click(functools.partial(button_move, dx=0, dy=-1))
    control_panel[5, 3] = bm1y

    bm5y = create_expanded_button("-5mm")
    bm5y.on_click(functools.partial(button_move, dx=0, dy=-5))
    control_panel[6, 3] = bm5y

    # x buttons
    bp5x = create_expanded_button("+5mm")
    bp5x.on_click(functools.partial(button_move, dx=5, dy=0))
    control_panel[3, 6] = bp5x

    bp1x = create_expanded_button("+1mm")
    bp1x.on_click(functools.partial(button_move, dx=1, dy=0))
    control_panel[3, 5] = bp1x

    bppt1x = create_expanded_button("+.1mm")
    bppt1x.on_click(functools.partial(button_move, dx=0.1, dy=0))
    control_panel[3, 4] = bppt1x

    bmpt1x = create_expanded_button("-.1mm")
    bmpt1x.on_click(functools.partial(button_move, dx=-0.1, dy=0))
    control_panel[3, 2] = bmpt1x

    bm1x = create_expanded_button("-1mm")
    bm1x.on_click(functools.partial(button_move, dx=-1, dy=0))
    control_panel[3, 1] = bm1x

    bm5x = create_expanded_button("-5mm")
    bm5x.on_click(functools.partial(button_move, dx=-5, dy=0))
    control_panel[3, 0] = bm5x

    return control_panel



import json
import os
import time
import webbrowser
from typing import Tuple, Union

import cv2
import matplotlib.pyplot as plt
import numpy as np
import requests

from science_jubilee.labware.Labware import Labware, Location, Well
from science_jubilee.tools.Tool import Tool, requires_active_tool


class Camera(Tool):
    """A class representation of a Raspberry Pi camera server client."""

    def __init__(
        self,
        index,
        name,
        ip_address,
        port,
        video_endpoint,
        still_endpoint,
        image_folder,
        focus_height,
        light: bool = False,
        light_pin: int = None,
    ):
        """Initializes the Camera object.

        :param index: The tool index of the pipette on the machine
        :type index: int
        :param name: The name associated with the tool (e.g. 'p300_single')
        :type name: str
        :param ip_address: The IP address of the Raspberry Pi camera server
        :type ip_address: str
        :param port: The port for the camera http requests
        :type port: int
        :param video_endpoint: The endpoint for the video feed
        :type video_endpoint: str
        :param still_endpoint: The endpoint for the still image
        :type still_endpoint: str
        :param image_folder: The folder to save captured images to
        :type image_folder: str
        :param light: LED ring light associated with the camera, defaults to False
        :type light: bool, optional
        :param light_pin: The GPIO pin (defined in the `M950` command in the machine `config.g`) for the LED ring light, defaults to None
        :type light_pin: int, optional
        """

        super().__init__(
            index,
            name,
            ip_address=ip_address,
            port=port,
            video_endpoint=video_endpoint,
            still_endpoint=still_endpoint,
            image_folder=image_folder,
            light=light,
            light_pin=light_pin,
            focus_height=focus_height,
        )
        self.still_url = f"http://{self.ip_address}:{self.port}/{self.still_endpoint}"
        self.video_url = f"http://{self.ip_address}:{self.port}/{self.video_endpoint}"

        # TODO: Ping camera server and make sure that it is reachable

    @classmethod
    def from_config(
        cls,
        index,
        name,
        config_file: str,
        path: str = os.path.join(os.path.dirname(__file__), "configs"),
    ):
        """Initialize the pipette object from a config file

        :param index: The tool index of the pipette on the machine
        :type index: int
        :param name: The name associated with the tool (e.g. 'WebCamera')
        :type name: str
        :param config_file: The name of the config file containing the tool parameters
        :type config_file: str
        :param path: The path to the labware configuration `.json` files for the labware,
                defaults to the 'labware_definition/' in the science_jubilee/labware directory.
        :type path: str, optional
        :return: the initialized :class:`Camera` object
        :rtype: :class:`Camera` object
        """

        config = os.path.join(path, config_file)
        with open(config, "rt") as f:
            kwargs = json.load(f)
        return cls(index=index, name=name, **kwargs)

    @requires_active_tool
    def _capture_image(self, timeout=30):
        """Capture image from raspberry pi and write to file

        :param timeout: the timeout for the http request, defaults to 10
        :type timeout: int, optional
        :return: the image as a bstring
        :rtype: bytes
        """
        time.sleep(1)
        try:
            response = requests.get(self.still_url, timeout=timeout)
        except [ConnectionError, ConnectionRefusedError]:
            raise AssertionError
        time.sleep(2)
        assert response.status_code == 200

        return response.content

    @requires_active_tool
    def capture_image(
        self,
        location: Union[Well, Tuple],
        light: bool = False,
        light_intensity: int = 0,
        timeout=30,
    ):
        """Capture an image from the WebCamera at the specified location

        :param location: the location of the well to capture an image of
        :type location: Union[Well, Tuple]
        :param light: Option to turn on a ring light before taking the image, defaults to False
        :type light: bool, optional
        :param light_intensity: Intensity of the led rign light, defaults to 0
        :type light_intensity: int, optional
        :return: the image as an bstring
        :rtype: bytes
        """
        assert 0 <= light_intensity <= 1, "Light intensity must be between 0 and 1"

        x, y, z = Labware._getxyz(location)

        self._machine.safe_z_movement()
        self._machine.move_to(x=x, y=y, wait=True)

        picture_heigth = self.focus_height - abs(self.tool_offset)
        self._machine.move_to(z=picture_heigth, wait=True)
        if light is True:
            self._machine.gcode(f"M42 P{self.light_pin} S{light_intensity}")
            image = self._capture_image(timeout=timeout)
            self._machine.gcode(f"M42 P{self.light_pin} S0")
        else:
            image = self._capture_image()

        return image

    def video_feed(self):
        """Deploys a video feed at the url specified in the config file"""
        webbrowser.open(self.video_url)

    def decode_image(self, image_bin):
        """Decode a bstring image into an np.array

        :param image_bin: the image as a bstring
        :type image_bin: bytes
        :return: the image as an np.array
        :rtype: np.array
        """
        image_arr = np.frombuffer(image_bin, np.uint8)
        image = cv2.imdecode(image_arr, cv2.IMREAD_COLOR)
        image_rgb = image[:, :, [2, 1, 0]]
        return image_rgb

    def process_image(self, image_bin, radius=50):
        """Externally callable function to run processing pipeline

        :param image_bin: the image as a bstring
        :type image_bin: bytes
        :param radius: the radius (in pixels) of the circular mask, defaults to 50
        :type radius: int, optional
        :return: the average rgb values of the masked image
        :rtype: list

        """
        image = self.decode_image(image_bin)
        r = radius
        masked_image = self._mask_image(image, r)
        t = time.time()
        cv2.imwrite(f"./sampleimage_full_{t}.jpg", image)
        cv2.imwrite(f"./sampleimage_masked_{t}.jpg", masked_image)
        rgb_values = self._get_rgb_avg(masked_image)
        return rgb_values

    def _mask_image(self, image, radius=50):
        """Apply a circular mask to an image

        :param image: the image object
        :type image: np.array
        :param radius: the size (in pixels) of the circular mask, defaults to 50
        :type radius: int, optional
        :return: the masked image
        :rtype: np.array
        """
        image_shape = image.shape[:2]
        w = image_shape[0] // 2
        h = image_shape[1] // 2
        mask = np.zeros(image_shape, dtype="uint8")
        cv2.circle(mask, (w, w), radius, 255, -1)
        masked = cv2.bitwise_and(image, image, mask=mask)

        return masked

    def _get_rgb_avg(self, image):
        """Extract the average rgb values from an image

        :param image: the image object
        :type image: np.array
        :return: the average rgb values in a list [R,G,B]
        :rtype: list
        """
        bgr = []
        for dim in [0, 1, 2]:
            flatdim = image[:, :, dim].flatten()
            indices = flatdim.nonzero()[0]
            value = flatdim.flatten()[indices].mean()
            bgr.append(value)

        # opencv uses bgr so convert to rgb for loss
        print("swapping")
        rgb = [bgr[i] for i in [2, 1, 0]]
        return rgb

    def view_image(self, image_bin, masked=False, radius=50):
        """Show the image in a matplotlib window

        :param image_bin: the image as a bstring
        :type image_bin: bytes
        :param masked: Wether to mask the image or not, defaults to False
        :type masked: bool, optional
        :param radius: the size (in pixel) of the circular mask toapply to the image , defaults to 50
        :type radius: int, optional
        """

        image = self.decode_image(image_bin)
        if masked is True:
            image = self._mask_image(image, radius)
        else:
            pass

        fig, ax = plt.subplots(figsize=(3, 4))
        plt.setp(plt.gca(), autoscale_on=True)
        ax.imshow(image)



"""Driver for Controlling Jubilee"""

# import websocket # for reading the machine model

import json
import os
import time
import warnings
from functools import wraps
from pathlib import Path
from typing import Union

import requests  # for issuing commands
from requests.adapters import HTTPAdapter, Retry

from science_jubilee.decks.Deck import Deck
from science_jubilee.tools.Tool import Tool

# TODO: Figure out how to print error messages from the Duet.


# copied from machine agency version, may not be needed here


def get_root_dir():
    """Return the path to the duckbot directory."""
    return Path(__file__).parent.parent


##########################################
#               ERRORS
##########################################
class MachineConfigurationError(Exception):
    """Raise this error if there is something wrong with how the machine is configured"""

    pass


class MachineStateError(Exception):
    """Raise this error if the machine is in the wrong state to perform the requested action."""

    pass


##########################################
#               DECORATORS
##########################################
def machine_homed(func):
    """Decorator used to check if the machine is homed before performing certain actions."""

    def homing_check(self, *args, **kwds):
        # Check the cached value if one exists.
        if self.simulated:
            return func(self, *args, **kwds)
        if self.axes_homed and all(self.axes_homed):
            return func(self, *args, **kwds)
        # Request homing status from the object model if not known.
        self.axes_homed = json.loads(self.gcode('M409 K"move.axes[].homed"'))["result"]
        if not all(self.axes_homed):
            raise MachineStateError("Error: machine must first be homed.")
        return func(self, *args, **kwds)

    return homing_check


def requires_deck(func):
    """Decorator used ot check if a deck has been configured before performing certain actions."""

    def deck_check(self, *args, **kwds):
        if self.deck is None:
            raise MachineStateError("Error: No deck is set up")
        return func(self, *args, **kwds)

    return deck_check


def requires_safe_z(func):
    """Decorator used to ensure the deck is at a safe height before performing certain actions."""

    def z_check(self, *args, **kwds):
        current_z = float(self.get_position()["Z"])
        if self.deck:
            safe_z = self.deck.safe_z
        else:
            safe_z = 0
            # warnings.warn(f"No deck configured, safe z height has been set to {safe_z}. Please modify this if needed.")
        if current_z < safe_z:
            self.move_to(z=safe_z + 20)
        return func(self, *args, **kwds)

    return z_check


##########################################
#             MACHINE CLASS
##########################################


class Machine:
    """A class representation of Jubilee used to send  motion commands and polling the machine state."""

    # TODO: Set this up so that a keyboard interrupt leaves the machine in a safe state - ie tool offsets correct. I had an issue
    # where I keyboard interrupted during pipette tip pickup - tip was picked up but offset was not applied, crashing machine on next move. This should not be possible.

    LOCALHOST = "192.168.1.2"

    def __init__(
        self,
        port: str = None,
        baudrate: int = 115200,
        address: str = None,
        deck_config: str = None,
        simulated: bool = False,
    ):
        """Initialize the Machine object.

        :param port: The port to connect to the machine over serial, defaults to None
        :type port: str, optional
        :param baudrate: The baudrate to use when connecting to the machine, defaults to 115200
        :type baudrate: int, optional
        :param address: The IP address of the machine. This should match what is loaded onto the config.g on the Jubilee Duet's main board, defaults to None
        :type address: str, optional
        :param deck_config: The name of the deck configuration file to load, defaults to None
        :type deck_config: str, optional
        :param simulated: Whether to simulate the machine, defaults to False
        :type simulated: bool, optional

        :raises MachineStateError: If the machine is not in the correct state to perform the requested action. This is a user error, not a machine error.
        :raises MachineConfigurationError: If the machine does nto support the indicated configuration, e.g., a tool index is already in use.
        :raises ValueError: If Jubilee returns an invalid value, e.g., the axis limit queried is not correct or query is unsuccessful.

        """
        if address != self.__class__.LOCALHOST:
            print(
                "Warning: disconnecting this application from the network will halt connection to Jubilee."
            )
        # Machine Specs

        # serial info
        self.ser = None
        self.port = port
        self.baudrate = baudrate
        self.lineEnding = "\n"  # serial stuff

        # HTTP info
        self.address = address

        # self.debug = debug
        self.simulated = simulated
        self.model_update_timestamp = 0
        self.command_ws = None
        self.wake_time = None  # Next scheduled time that the update thread updates.

        self._absolute_positioning = True
        self._absolute_extrusion = (
            True  # Extrusion positioning is set separately from other axes
        )
        self._configured_axes = None
        self._configured_tools = None
        self._active_tool_index = None  # Cached value under the @property.
        self._tool_z_offsets = None  # Cached value under the @property.
        self._axis_limits = (None, None, None)  # Cached value under the @property.
        self.axes_homed = [
            False
        ] * 4  # We have at least X/Y/Z/U axes to home. Additional axes handled below in connect()
        self.deck = None
        # TODO: this is confusingly named
        self.tools = {}  # this is the list of available tools
        self.tool = None  # this is the current active tool
        self.current_well = None

        requests_session = requests.Session()
        retries = Retry(
            total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504]
        )

        requests_session.mount("http://", HTTPAdapter(max_retries=retries))
        self.session = requests_session

        if deck_config is not None:
            self.load_deck(deck_config)

        self.connect()

        self._set_absolute_positioning()  # force=True)

    def connect(self):
        """Connects to Jubilee over http.

        :raises MachineStateError: If the connection to the machine is unsuccessful.
        """
        # TODO: incorporate serial connection from machine agency version
        if self.simulated:
            return
        # Do the equivalent of a ping to see if the machine is up.

        # if self.debug:
        #    print(f"Connecting to {self.address} ...")
        try:
            # "Ping" the machine by updating the only cacheable information we care about.
            # TODO: This should handle a response from self.gcode of 'None' gracefully.
            max_tries = 50
            for i in range(max_tries):
                response = json.loads(self.gcode('M409 K"move.axes[].homed"'))[
                    "result"
                ][:4]
                if len(response) == 0:
                    continue
                else:
                    break
            self.axes_homed = response

            # These data members are tied to @properties of the same name
            # without the '_' prefix.
            # Upon reconnecting, we need to flag that the @property must
            # refresh; otherwise we will retrieve old values that may be invalid.
            self._active_tool_index = None
            self._tool_z_offsets = None
            self._axis_limits = None

            # To save time upon connecting, let's just hit the API on the
            # first try for all the @properties we care about.
            self.configured_axes
            self.active_tool_index
            self.tool_z_offsets
            self.axis_limits
            # pprint.pprint(json.loads(requests.get("http://127.0.0.1/machine/status").text))
            # TODO: recover absolute/relative from object model instead of enforcing it here.
            self._set_absolute_positioning()
        except json.decoder.JSONDecodeError as e:
            raise MachineStateError("DCS not ready to connect.") from e
        except requests.exceptions.Timeout as e:
            raise MachineStateError(
                "Connection timed out. URL may be invalid, or machine may not be connected to the network."
            ) from e
        # if self.debug:
        #    print("Connected.")

    @property
    def configured_axes(self):
        """Return the configured axes of the machine.
        This is a list of all configured axis on the machine. Usually these are ['X', 'Y', 'Z', 'U']. If a tool has a
        an associated driver, it will be appended at the end of the axis list.

        :return: A list of configured axes.
        :rtype: list
        """
        if self._configured_axes is None:  # Starting from a fresh connection
            try:
                max_tries = 50
                for i in range(max_tries):
                    response = json.loads(self.gcode('M409 K"move.axes[]"'))["result"]
                    if len(response) == 0:
                        continue
                    else:
                        break
                self._configured_axes = []
                for axis in response:
                    self._configured_axes.append(axis["letter"])
            except ValueError as e:
                print("Error occurred trying to read axis limits on each axis!")
                raise e

        # Return the cached value.
        return self._configured_axes

    @property
    def configured_tools(self):
        """Returns a list of configured tools.

        :return: A list of configured tools.The tool name is queried from the `config.g` file on the machine.
        :rtype: list

        Note: This list is obtained directly from the tools added to the machine's `config.g` file.
        """

        # TODO: Compare this to loaded tools list
        if self._configured_tools is None:  # Starting from a fresh connection
            try:
                max_tries = 50
                for i in range(max_tries):
                    response = json.loads(self.gcode('M409 K"tools[]"'))["result"]
                    if len(response) == 0:
                        continue
                    else:
                        break
                self._configured_tools = {}
                for tool in response:
                    self._configured_tools[tool["number"]] = tool["name"]
            except ValueError as e:
                print("Error occurred trying to read axis limits on each axis!")
                raise e

        # Return the cached value.
        return self._configured_tools

    @property
    def active_tool_index(self):
        """Return the index of the current active tool.

        :return: The index of the current active tool. If no tool is active, returns -1.
        :rtype: int
        """

        if self._active_tool_index is None:  # Starting from a fresh connection.
            try:
                max_tries = 50
                for i in range(max_tries):
                    response = self.gcode("T")
                    if len(response) == 0:
                        continue
                    else:
                        break
                # On HTTP Interface, we get a string instead of -1 when there are no tools.
                if response.startswith("No tool"):
                    # print('active tool prop thinks theres no tool')
                    return -1
                # On HTTP Interface, we get a string instead of the tool index.
                elif response.startswith("Tool"):
                    # Recover from the string: 'Tool X is selected.'
                    self.active_tool_index = int(response.split()[1])
                else:
                    self.active_tool_index = int(response)
            except ValueError as e:
                # print("Error occurred trying to read current tool!")
                raise e
        # Return the cached value.
        return self._active_tool_index

    @active_tool_index.setter
    def active_tool_index(self, tool_index: int):
        """Sets the current tool, and toggle the old tool off."""
        if self.tool is not None:
            self.tool.is_active_tool = False

        if tool_index < 0:
            self._active_tool_index = -1
            self.tool = None
        else:
            self._active_tool_index = tool_index
            if tool_index not in self.tools:
                temp_tool = Tool(tool_index, "temp_tool")
                self.load_tool(temp_tool)
            tool = self.tools[tool_index]["tool"]
            self.tool = tool
            tool.is_active_tool = True

    @property
    def tool_z_offsets(self):
        """Return (in tool order) a list of tool's z offsets.

        Note: This list is obtained directly from the tools added to the machine's `config.g` file.

        :return: A list of tool z offsets, in the order of the tool index
        :rtype: list"""
        # Starting from fresh connection, query from the Duet.
        # if self._tool_z_offsets is None:
        try:
            max_tries = 50
            for i in range(max_tries):
                response = json.loads(self.gcode('M409 K"tools"'))["result"]
                if len(response) == 0:
                    continue
                else:
                    break

            self._tool_z_offsets = {}  # Create a fresh dictionary.
            for tool_data in response:
                if tool_data is None:
                    continue
                tool_number = tool_data["number"]
                tool_z_offset = tool_data["offsets"][2]  # Pull Z axis
                self._tool_z_offsets[tool_number] = tool_z_offset
        except ValueError as e:
            print("Error occurred trying to read z offsets of all tools!")
            raise e
        # Return the cached value.
        return self._tool_z_offsets

    @property
    def axis_limits(self):
        """Return (in XYZU order) a list of tuples specifying (min, max) axis limit

        :return: A list of tuples specifying (min, max) axis limit
        :rtype: list

        Note: This list is obtained directly from the tools added to the machine's `config.g` file.
        """
        # Starting from fresh connection, query from the Duet.
        if self._axis_limits is None:
            try:
                max_tries = 50
                for i in range(max_tries):
                    response = json.loads(self.gcode('M409 K"move.axes"'))["result"]
                    if len(response) == 0:
                        continue
                    else:
                        break
                # pprint.pprint(response)
                self._axis_limits = []  # Create a fresh list.
                for axis_data in response:
                    axis_min = axis_data["min"]
                    axis_max = axis_data["max"]
                    self._axis_limits.append((axis_min, axis_max))
            except ValueError as e:
                print("Error occurred trying to read axis limits on each axis!")
                raise e
        # Return the cached value.
        return self._axis_limits

    @property
    def position(self):
        """Returns the current machine control point in mm.

        :return: A dictionary of the machine control point in mm. The keys are the axis name, e.g. 'X'
        :rtype: dict
        """
        # Axes are ordered X, Y, Z, U, E, E0, E1, ... En, where E is a copy of E0.
        response_chunks = self.gcode("M114").split()
        positions = [float(a.split(":")[1]) for a in response_chunks[:3]]
        return positions

    ##########################################
    #                BED PLATE
    ##########################################
    def load_deck(
        self,
        deck_filename: str,
        path: str = os.path.join(os.path.dirname(__file__), "decks", "deck_definition"),
    ):
        """Load a deck configuration file onto the machine.

        :param deck_filename: The name of the deck configuration file.
        :type deck_filename: str
        :param path: The path to the deck configuration `.json` files for the labware,
                defaults to the 'deck_definition/' in the science_jubilee/decks directory.
        :type path: str, optional
        :return: A :class:`Deck` object
        :rtype: :class:`Deck`
        """
        deck = Deck(deck_filename, path=path)
        self.deck = deck
        return deck

    def gcode(self, cmd: str = "", timeout=None, response_wait: float = 30):
        """Send a G-Code command to the Machine and return the response.

        :param cmd: The G-Code command to send, defaults to ""
        :type cmd: str, optional
        :param timeout: The time to wait for a response from the machine, defaults to None
        :type timeout: float, optional
        :param response_wait: The time to wait for a response from the machine, defaults to 30
        :type response_wait: float, optional

        :return: The response message from the machine. If too long, the message might not display in the terminal.
        :rtype: str
        """

        # TODO: Add serial option for gcode commands from MA
        if self.simulated:
            print(f"sending: {cmd}")
            return None

        try:
            # Try sending the command with requests.post
            response = requests.post(
                f"http://{self.address}/machine/code", data=f"{cmd}", timeout=timeout
            ).text
            if "rejected" in response:
                raise requests.RequestException
        except requests.RequestException:
            # If requests.post fails ( not supported for standalone mode), try sending the command with requests.get
            try:
                # Paraphrased from Duet HTTP-requests page:
                # Client should query `rr_model?key=seqs` and monitor `seqs.reply`. If incremented, the command went through
                # and the response is available at `rr_reply`.
                reply_count = self.session.get(
                    f"http://{self.address}/rr_model?key=seqs"
                ).json()["result"]["reply"]
                buffer_response = self.session.get(
                    f"http://{self.address}/rr_gcode?gcode={cmd}", timeout=timeout
                )
                # wait for a response code to be appended
                # TODO: Implement retry backoff for managing long-running operations to avoid too many requests error. Right now this is handled by the generic exception catch then sleep. Real fix is some sort of backoff for things running longer than a few seconds.
                tic = time.time()
                while True:
                    try:
                        new_reply_count = self.session.get(
                            f"http://{self.address}/rr_model?key=seqs"
                        ).json()["result"]["reply"]
                        if new_reply_count != reply_count:
                            response = self.session.get(
                                f"http://{self.address}/rr_reply"
                            ).text
                            break
                        elif time.time() - tic > response_wait:
                            response = None
                            break
                    except Exception as e:
                        print("Connection error, sleeping 1 second")
                        time.sleep(2)
                        continue

            except requests.RequestException as e:
                print(f"Both `requests.post` and `requests.get` requests failed: {e}")
                response = None
        # TODO: handle this with logging. Also fix so all output goes to logs
        return response

    def _set_absolute_positioning(self):
        """Set absolute positioning for all axes except extrusion"""
        self.gcode("G90")
        self._absolute_positioning = True

    def _set_relative_positioning(self):
        """Set relative positioning for all axes except extrusion"""
        self.gcode("G91")
        self.absolute_positioning = False

    def _set_absolute_extrusion(self):
        """Set absolute positioning for extrusion"""
        self.gcode("M82")
        self._absolute_extrusion = True

    def _set_relative_extrusion(self):
        """Set relative positioning for extrusion"""
        self.gcode("M83")
        self.absolute_extrusion = False

    def push_machine_state(self):
        """Push machine state onto a stack"""
        self.gcode("M120")

    def pop_machine_state(self):
        """Recover previous machine state"""
        self.gcode("M121")

    def download_file(self, filepath: str = None, timeout: float = None):
        """Download a file into a file object. Full machine filepath must be specified.
        Example: /sys/tfree0.g

        :param filepath: The full filepath of the file to download, defaults to None
        :type filepath: str, optional
        :param timeout: The time to wait for a response from the machine, defaults to None
        :type timeout: float, optional
        :return: The file contents
        :rtype: file object
        """
        # RRF3 Only
        file_contents = requests.get(
            f"http://{self.address}/rr_download?name={filepath}", timeout=timeout
        )
        return file_contents

    def reset(self):
        """Issue a software reset."""
        # End the subscribe thread first.
        self.gcode("M999")  # Issue a board reset. Assumes we are already connected
        self.axes_homed = [False] * 4
        self.disconnect()
        print("Reconnecting...")
        for i in range(10):
            time.sleep(1)
            try:
                self.connect()
                return
            except MachineStateError as e:
                pass
        raise MachineStateError("Reconnecting failed.")

    def home_all(self):
        """Home all axes."""
        # Having a tool is only possible if the machine was already homed.
        # TODO: Check if machine is already homed and have a user input to verify clear deck to avoid wasting time by accidentally rerunning and \
        # avoid major deck wrecks
        # TODO: Catch errors where tool is already on and forward to user for fix
        if self.active_tool_index != -1:
            self.park_tool()
        self.gcode("G28")
        self._set_absolute_positioning()
        # Update homing state. Do not query the object model because of race condition.
        self.axes_homed = [True, True, True, True]  # X, Y, Z, U

        ### test to see if we can get the number of axis home using the pop_machine_state(self) !! MP 07/25/23

    def home_xyu(self):
        """Home the XYU axes. Home Y before X to prevent possibility of crashing into the tool rack."""
        self.gcode("G28 Y")
        self.gcode("G28 X")
        self.gcode("G28 U")
        self._set_absolute_positioning()
        # Update homing state. Pull Z from the object model which will not create a race condition.
        z_home_status = json.loads(self.gcode('M409 K"move.axes[].homed"'))["result"][2]
        self.axes_homed = [True, True, z_home_status, True]

    def home_x(self):
        """Home the X axis"""
        cmd = "G28 X"
        self.gcode(cmd)

    def home_y(self):
        """Home the Y axis"""
        cmd = "G28 Y"
        self.gcode(cmd)

    def home_u(self):
        """Home the U (tool) axis"""
        cmd = "G28 U"
        self.gcode(cmd)

    def home_v(self):
        """Home the V axis"""
        cmd = "G28 V"
        self.gcode(cmd)

    def home_z(self):
        """Home the Z axis.
        Note: The deck must be clear first. Will ask for user input to verify.
        """
        response = input("Is the Deck free of obstacles? [y/n]")
        if response.lower() in ["y", "yes", "Yes", "Y", "YES"]:
            self.gcode("G28 Z")
        else:
            print("The deck needs to be empty of all labware before proceeding.")
        self._set_absolute_positioning()

    def home_e(self):
        """
        Home the extruder axis (syringe)
        """
        pass

    def home_in_place(self, *args: str):
        """Set the current location of a machine axis or axes to 0."""
        for axis in args:
            if axis.upper() not in ["X", "Y", "Z", "U"]:
                raise TypeError(f"Error: cannot home unknown axis: {axis}.")
            self.gcode(f"G92 {axis.upper()}0")

    def set_tool_offset(self, tool_idx=None, x=None, y=None, z=None):
        if tool_idx is None:
            raise MachineConfigurationError("No tool index provided!")

        x = "{0:.2f}".format(x) if x is not None else None
        y = "{0:.2f}".format(y) if y is not None else None
        z = "{0:.2f}".format(z) if z is not None else None

        p_cmd = x_cmd = y_cmd = z_cmd = ""

        if tool_idx is not None:
            p_cmd = f"P{tool_idx}"
        if x is not None:
            x_cmd = f"X{x}"
        if y is not None:
            y_cmd = f"Y{y}"
        if z is not None:
            z_cmd = f"Z{z}"

        cmd = f"G10 {p_cmd} {z_cmd} {x_cmd} {y_cmd}"
        self.gcode(cmd)

    @machine_homed
    def _move_xyzev(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        e: float = None,
        v: float = None,
        s: float = 6000,
        param: str = None,
        wait: bool = False,
    ):
        """Move X/Y/Z/E/V axes. Set absolute/relative mode externally.

        :param x: x position on the bed, in whatever units have been set (default mm)
        :type x: float, optional
        :param y: y position on the bed, in whatever units have been set (default mm)
        :type y: float, optional
        :param z: z position on the bed, in whatever units have been set (default mm)
        :type z: float, optional
        :param e: extruder position, in whatever units have been set (default mm)
        :type e: float, optional
        :param v: v axis position, in whatever units have been set (default mm)
        :type v: float, optional
        :param s: speed at which to move (default 6000 mm/min)
        :type s: float, optional
        """

        x = "{0:.2f}".format(x) if x is not None else None
        y = "{0:.2f}".format(y) if y is not None else None
        z = "{0:.2f}".format(z) if z is not None else None
        e = "{0:.2f}".format(e) if e is not None else None
        v = "{0:.2f}".format(v) if v is not None else None
        s = "{0:.2f}".format(s)

        # initialize coordinates commands
        x_cmd = y_cmd = z_cmd = e_cmd = v_cmd = f_cmd = param_cmd = ""

        if x is not None:
            x_cmd = f"X{x}"
        if y is not None:
            y_cmd = f"Y{y}"
        if z is not None:
            z_cmd = f"Z{z}"
        if e is not None:
            e_cmd = f"E{e}"
        if v is not None:
            v_cmd = f"V{v}"
        if s is not None:
            f_cmd = f"F{s}"
        if param is not None:
            param_cmd = param

        cmd = f"G0 {z_cmd} {x_cmd} {y_cmd} {e_cmd} {v_cmd} {f_cmd} {param_cmd}"
        self.gcode(cmd)
        if wait:
            self.gcode(f"M400")

    def move_to(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        e: float = None,
        v: float = None,
        s: float = 6000,
        param: str = None,
        wait: bool = False,
    ):
        """Move to an absolute X/Y/Z/E/V position.

        :param x: x position on the bed, in whatever units have been set (default mm)
        :type x: float, optional
        :param y: y position on the bed, in whatever units have been set (default mm)
        :type y: float, optional
        :param z: z position on the bed, in whatever units have been set (default mm)
        :type z: float, optional
        :param e: extruder position, in whatever units have been set (default mm)
        :type e: float, optional
        :param v: v axis position, in whatever units have been set (default mm)
        :type v: float, optional
        :param s: speed at which to move (default 6000 mm/min)
        :type s: float, optional

        """
        self._set_absolute_positioning()

        self._move_xyzev(x=x, y=y, z=z, e=e, v=v, s=s, param=param, wait=wait)

    def move(
        self,
        dx: float = 0,
        dy: float = 0,
        dz: float = 0,
        de: float = 0,
        dv: float = 0,
        s: float = 6000,
        param: str = None,
        wait: bool = False,
    ):
        """Move relative to the current position

        :param dx: change in x position, in whatever units have been set (default mm)
        :type dx: float, optional
        :param dy: change in y position, in whatever units have been set (default mm)
        :type dy: float, optional
        :param dz: change in z position, in whatever units have been set (default mm)
        :type dz: float, optional
        :param de: change in e position, in whatever units have been set (default mm)
        :type de: float, optional
        :param dv: change in v position, in whatever units have been set (default mm)
        :type dv: float, optional
        :param s:  speed at which to move (default 6000 mm/min)
        :type s: float, optional
        """
        # Check that the relative move doesn't exceed user-defined limit
        # By default, ensure that it won't crash into the parked tools
        if any(self._axis_limits):
            x_limit, y_limit, z_limit = self._axis_limits[0:3]
            pos = self.get_position()
            if (
                x_limit
                and dx != 0
                and (
                    (float(pos["X"]) + dx > x_limit[1])
                    or (float(pos["X"]) + dx < x_limit[0])
                )
            ):
                raise MachineStateError("Error: Relative move exceeds X axis limit!")
            if (
                y_limit
                and dy != 0
                and (
                    (float(pos["Y"]) + dy > y_limit[1])
                    or (float(pos["Y"]) + dy < y_limit[0])
                )
            ):
                raise MachineStateError("Error: Relative move exceeds Y axis limit!")
            if (
                z_limit
                and dz != 0
                and (
                    (float(pos["Z"]) + dz > z_limit[1])
                    or (float(pos["Z"]) + dz < z_limit[0])
                )
            ):
                raise MachineStateError("Error: Relative move exceeds Z axis limit!")
        self._set_relative_positioning()

        self._move_xyzev(x=dx, y=dy, z=dz, e=de, v=dv, s=s, param=param, wait=wait)

    def dwell(self, t: float, millis: bool = True):
        """Pauses the machine for a period of time.

        :param t: time to pause, in milliseconds by default
        :type t: float
        :param millis: boolean, set to false to use seconds. default unit is milliseconds.
        :type millis: bool, optional
        """

        param = "P" if millis else "S"
        cmd = f"G4 {param}{t}"

        self.gcode(cmd)

    def safe_z_movement(self):
        """Move the Z axis to a safe height to avoid crashing into labware."""
        # TODO is this redundant? can we reuse decorator ?
        current_z = self.get_position()["Z"]
        safe_z = self.deck.safe_z
        if float(current_z) < safe_z:
            self.move_to(z=safe_z + 20)
        else:
            pass

    def _get_tool_index(self, tool_item: Union[int, Tool, str]):
        """Return the tool index from the provided tool item.

        This method is allows the user to call a toll by its index, its name, or to use a :class:`Tool` object
        directly.

        :param tool_item: The tool index, name, or :class:`Tool` object
        :type tool_item: Union[int, Tool, str]
        :return: The tool index
        :rtype: int
        """
        if type(tool_item) == int:
            assert tool_item in set(self.tools.values()), f"Tool {tool_item} not loaded"
            return tool_item
        elif type(tool_item) == str:
            assert tool_item in set(self.tools.values()), f"Tool {tool_item} not loaded"
            return self.tools[tool_item]
        elif isinstance(tool_item, Tool):
            assert tool_item.index in set(
                self.tools.keys()
            ), f"Tool {tool_item} not loaded"
            return tool_item.index
        else:
            raise ValueError(f"Unknown tool format {type(tool_item)}")

    def load_tool(self, tool: Tool = None):
        """Add a new tool for use on the machine."""
        # TODO: Fix this so if you reload you don't break everything
        name = tool.name
        idx = tool.index

        # Ensure that the provided tool index and name are unique.
        if idx in self.tools:
            # Handle the case that connection was established with a tool equipped
            if self.tools[idx]["name"] == "temp_tool":
                tool.is_active_tool = True
        for loaded_tool in self.tools.values():
            if loaded_tool["name"] is name and loaded_tool["tool"].index != idx:
                raise MachineConfigurationError("Error: Tool name already in use.")

        self.tools[idx] = {"name": name, "tool": tool}
        tool._machine = self
        tool.post_load()
        tool.tool_offset = self.tool_z_offsets[idx]

    def reload_tool(self, tool: Tool = None):
        """Update a tool which has already been loaded."""
        name = tool.name
        idx = tool.index

        # Ensure that the provided tool index and name are unique.
        if idx not in self.tools:
            raise MachineConfigurationError(
                f"Error: No tool with index {idx} to update."
            )
        for loaded_tool in self.tools.values():
            if loaded_tool["name"] is name:
                raise MachineConfigurationError("Error: Tool name already in use.")

        self.tools[idx] = {"name": name, "tool": tool}
        tool._machine = self

    # TODO: Unload tool method

    @requires_safe_z
    def pickup_tool(self, tool_id: Union[int, str, Tool]):
        """Pick up the tool specified by a tool index, name or :class:`Tool` object.

        :param tool_id: The tool index, name, or :class:`Tool` object
        :type tool_id: Union[int, str, Tool]
        :raises MachineConfigurationError: If the tool is not loaded on the machine.
        :raises ValueError: If the indicated tool_id is not of type Union[int, str, Tool].
        """
        # TODO: Make sure axis limits are checked and not exceeded when picking up pipette
        if isinstance(
            tool_id, int
        ):  # Accept either tool index, tool name, or reference to the tool itself
            if tool_id in self.tools:
                tool_index = tool_id
            else:
                raise MachineConfigurationError(
                    f"Error: No tool with index {tool_id} is currently loaded."
                )
        elif isinstance(tool_id, str):
            tool_index = None
            for loaded_tool_index in self.tools:
                if self.tools[loaded_tool_index]["name"] is tool_id:
                    tool_index = loaded_tool_index
                    break
            if tool_index is None:
                raise MachineConfigurationError(
                    f"Error: No tool with name {tool_id} is currently loaded."
                )
        elif isinstance(tool_id, Tool):
            tool_index = None
            for loaded_tool_index in self.tools:
                if self.tools[loaded_tool_index]["tool"] is tool_id:
                    tool_index = loaded_tool_index
                    break
            if tool_index is None:
                raise MachineConfigurationError(
                    f"Error: No tool of type {tool_id} is currently loaded."
                )
        else:
            raise ValueError(f"Unknown tool format {type(tool_id)}")

        #         self.safe_z_movement()
        self.gcode(f"T{tool_index}")
        self.active_tool_index = tool_index
        self.tools[tool_index]["tool"].is_active_tool = True

    @requires_safe_z
    def park_tool(self):
        """Park the current tool adn cahnges active tool index to `-1`."""
        # self.safe_z_movement()
        self.gcode("T-1")
        # Update the cached value to prevent read delays.
        current_tool_index = self.active_tool_index
        self.tools[current_tool_index]["tool"].is_active_tool = False
        self._active_tool_index = -1

    def get_position(self):
        """Get the current position of the machine control point in mm.

        :return: A dictionary of the machine control point in mm. The keys are the axis name, e.g. 'X'
        :rtype: dict
        """

        max_tries = 50
        for i in range(max_tries):
            resp = self.gcode("M114")
            if "Count" not in resp:
                continue
            else:
                break
        positions = {}
        keyword = " Count "  # this is the keyword hosts like e.g. pronterface search for to track position
        keyword_idx = resp.find(keyword)

        count = 0
        if keyword_idx > -1:
            resp = resp[:keyword_idx]
            position_elements = resp.split(" ")
            for e in position_elements:
                axis, pos = e.split(":", 2)
                positions[axis] = pos

        return positions

    def load_labware(
        self, labware_filename: str, slot: int, path: str = None, order: str = "rows"
    ):
        """Function that loads a labware and associates it with a specific slot on the deck.
         The slot offset is also applied to the labware asocaite with it.

        :param labware_filename: The name of the labware configuration file.
        :type labware_filename: str
        :param slot: The index of the slot to load the labware into.
        :type slot: int
        :param path: The path to the labware configuration `.json` files for the labware.
        :type path: str, optional
        :param order: The order in which the labware is arranged on the deck.
                Can be 'rows' or 'columns', defaults to 'rows'.
        :type order: str, optional
        :return: The :class:`Labware` object that has been loaded into the slot.
        :rtype: :class:`Labware`
        """
        if path is not None:
            labware = self.deck.load_labware(
                labware_filename, slot, path=path, order=order
            )
        else:
            labware = self.deck.load_labware(labware_filename, slot, order=order)

        return labware

    # ***************MACROS***************
    def tool_lock(self):
        """Runs Jubilee tool lock macro. Assumes tool_lock.g macro exists."""
        cmd = 'M98 P"0:/macros/tool_lock.g"'
        self.gcode(cmd)

    def tool_unlock(self):
        """Runs Jubilee tool unlock macro. Assumes tool_unlock.g macro exists."""
        cmd = 'M98 P"0:/macros/tool_unlock.g"'
        self.gcode(cmd)

    def disconnect(self):
        """Close the connection."""
        # Nothing to do?
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.disconnect()



class ToolStateError(Exception):
    """Raise this error if the tool is in the wrong state to perform such a command."""

    pass


class ToolConfigurationError(Exception):
    """Raise this error if there is something wrong with how the tool is configured"""

    pass


class Tool:
    # TODO: Is this init supposed to take a machine?
    def __init__(self, index, name, **kwargs):
        if not isinstance(index, int) or not isinstance(name, str):
            raise ToolConfigurationError(
                "Incorrect usage: load_tool(<tool_number>, <name>, **kwargs)"
            )
        self._machine = None
        self.index = index
        self.name = name
        self.is_active_tool = False
        self.tool_offset = None

        for k, v in kwargs.items():
            setattr(self, k, v)

    def post_load(self):
        """Run any code after tool has been associated with the machine."""
        pass

    # TODO:
    # add a park tool method that every tool config can define to do things that need to be done pre or post parking
    # ex: make sure pipette has dropped tips before parking


def requires_active_tool(func):
    """Decorator to ensure that a tool cannot complete an action unless it is the
    current active tool.
    """

    def wrapper(self, *args, **kwargs):
        if self.is_active_tool == False:
            raise ToolStateError(
                f"Error: Tool {self.name} is not the current `Active Tool`. Cannot perform this action"
            )
        else:
            return func(self, *args, **kwargs)

    return wrapper



# # Iterate through and import all modules in the plates/
# # this allows us to add new plates in the plates folder and have it automatically usable in Machine.py
# from inspect import isclass
# from pkgutil import iter_modules
# from pathlib import Path
# from importlib import import_module

# # iterate through the modules in the current package
# package_dir = Path(__file__).resolve().parent
# for (_, module_name, _) in iter_modules([package_dir]):
#     # import the module and iterate through its attributes
#     module = import_module(f"{__name__}.{module_name}")
#     for attribute_name in dir(module):
#         attribute = getattr(module, attribute_name)

#         if isclass(attribute):
#             # Add the class to this package's variables
#             globals()[attribute_name] = attribute



import logging
import time
from typing import List, Tuple, Union

import serial

from science_jubilee.labware.Labware import Labware, Location, Well
from science_jubilee.tools.Tool import Tool, requires_active_tool

logger = logging.getLogger(__name__)


class Sonicator(Tool):
    """A class representation for a Qsonica Sonicator tool."""

    def __init__(self, index, name, mode: str, port: str = None):
        """Initialize a Sonicator object.

        :param index: index of the tool
        :type index: int
        :param name: name of the tool
        :type name: str
        :param mode: mode of the sonicator. Either Raspberry Pi 'Pico' or 'Hat'
        :type mode: str
        :param port: if using a Raspberry Pi pico, the usb port used to communicate with the Pico microcontroller need to be provided, this is computer dependent.
        :type port: str
        """
        super().__init__(index, name)
        assert mode in ["Pico", "Hat"], (
            "Error: mode must be either 'Pico' if using a Raspberry Pi Pico microcontroller"
            "or 'Hat' if using the custom pcb board hat for a Raspberry Pi SBC."
        )

        self.interface_mode = mode  # either Raspberry Pi Pico or Raspberry Pi Hat
        if mode == "Pico":
            assert (
                port is not None
            ), "Error: port must be specified for serial interface with Pico microcontroller"
            self._raspberrypi_pico(port)
        elif mode == "Hat":
            # initialize board pins
            self._raspberrypi_hat()
            pass

    def _raspberrypi_hat(self):
        """Initialize the pins used on the Raspberry Pi Hat."""

        import adafruit_mcp4725
        import board
        import busio
        import digitalio

        # define DAC pins for I2C communication
        self.i2c = busio.I2C(board.SCL, board.SDA)
        self.dac = adafruit_mcp4725.MCP4725(self.i2c, address=0x60)
        self.sonicator_enable = digitalio.DigitalInOut(board.D4)
        self.sonicator_enable.direction = digitalio.Direction.OUTPUT
        self.sonicator_enable.value = (
            False  # turns off D4 pin to ensure sonicator is off
        )
        self.dac.normalized_value = 0.0  # sets DAC to 0.0 to ensure sonicator is off

    def _raspberrypi_pico(self, port: str):
        """Initialize serial interface with Raspberry Pi Pico microcontroller.

        :param port: port to connect to the Pico microcontroller, this is computer dependent.
        :type port: str
        """
        ser = serial.Serial(port, baudrate=115200, timeout=1)
        ser.flush()  # let's clean buffer before we start communication
        self.serial_interface = ser

    def _set_sonication_power(self, power: float):
        """Set the power of the sonicator to a value between 0.4 and 1.0.

        :param power: power level to set the sonicator to. Must be between 0.4 and 1.0.
        :type power: float
        """
        if self.interface_mode == "Pico":
            self.serial_interface.write(f"DAC:{power}\n".encode("utf-8"))
            while self.serial_interface.out_waiting > 0:  # wait for message to be sent
                pass
        elif self.interface_mode == "Hat":
            self.dac.normalized_value = power

    @requires_active_tool
    def _sonication_on(self):
        """Turn on the sonicator."""
        if self.interface_mode == "Pico":
            self.serial_interface.write(b"SONICATOR_ON\n")
            while self.serial_interface.out_waiting > 0:
                pass
        elif self.interface_mode == "Hat":
            self.sonicator_enable.value = True

    def _sonication_off(self):
        """Turn off the sonicator."""
        if self.interface_mode == "Pico":
            self.serial_interface.write(b"SONICATOR_OFF\n")
            while self.serial_interface.out_waiting > 0:
                pass
        elif self.interface_mode == "Hat":
            self.sonicator_enable.value = False

    def _sonicate(
        self,
        exposure_time: float = 1.0,
        power: float = 0.4,
        pulse_duty_cycle: float = 0.5,
        pulse_interval: float = 1.0,
        verbose: bool = False,
    ):
        """Enable the sonicator at the power level for the exposure time.

        :param exposure_time: total time to sonicate for, defaults to 1.0
        :type exposure_time: float, optional
        :param power: power level to sonicate at, defaults to 0.4
        :type power: float, optional
        :param pulse_duty_cycle: duty cycle of the sonicator pulse, defaults to 0.5
        :type pulse_duty_cycle: float, optional
        :param pulse_interval: interval between pulses, defaults to 1.0
        :type pulse_interval: float, optional
        :param verbose: whether or not to print out the time elapsed, defaults to False
        :type verbose: bool, optional
        """
        # Quick sanity checks
        assert (
            0 <= power <= 1.0
        ), f"Error: power must be between 0.0 and 1.0. Value specified is: {power}"
        assert (
            0 <= pulse_duty_cycle <= 1.0
        ), f"Error: pulse_duty_cycle must be between 0.0 and 1.0. Value specified is: {pulse_duty_cycle}"
        assert (
            pulse_interval > 0
        ), f"Error: pulse_interval must be positive. Value specified is: {pulse_interval}."
        assert pulse_interval <= exposure_time, (
            f"Error: pulse_interval cannot exceed exposure time. Value specified is: {pulse_interval}, "
            f"but total exposure time is {exposure_time}."
        )

        # Set the power level for the sonicator through the DAC.
        self._set_sonication_power(power)
        on_interval = pulse_duty_cycle * pulse_interval
        off_interval = (1 - pulse_duty_cycle) * pulse_interval

        start_time = time.perf_counter()
        stop_time = exposure_time + start_time
        while True:
            # On interval.
            curr_time = time.perf_counter()
            if curr_time + on_interval < stop_time:
                if verbose:
                    print(f"{time.perf_counter() - start_time :.2f} | Sonicator on.")
                else:
                    pass
                self._sonication_on()
                time.sleep(on_interval)
            elif stop_time > curr_time:  # last time to sleep.
                if verbose:
                    print(f"{time.perf_counter() - start_time :.2f} | Sonicator on.")
                else:
                    pass
                self._sonication_on()
                time.sleep(stop_time - curr_time)

            # Off interval.
            curr_time = time.perf_counter()
            if curr_time + off_interval < stop_time:
                if verbose:
                    print(f"{time.perf_counter() - start_time :.2f} | Sonicator off.")
                else:
                    pass
                self._sonication_off()
                time.sleep(off_interval)
            elif stop_time > curr_time:  # last time to sleep.
                if verbose:
                    print(f"{time.perf_counter() - start_time :.2f} | Sonicator off.")
                else:
                    pass
                self._sonication_off()
                time.sleep(stop_time - curr_time)
                break
            else:
                if verbose:
                    print(f"{time.perf_counter() - start_time :.2f} | Sonicator off.")
                else:
                    pass
                break

        print(f"{time.perf_counter() - start_time :.2f} | Finished sonicating.")
        self._sonication_off()
        self._set_sonication_power(0.0)

    @requires_active_tool
    def sonicate_well(
        self,
        location: Union[Well, Tuple, Location],
        plunge_depth: float,
        sonication_time: float,
        power: float = 0.4,
        pulse_duty_cycle: float = 0.5,
        pulse_interval: float = 1.0,
        verbose: bool = False,
        autoclean: bool = False,
        *args,
    ):
        """Sonicate one well at a specified depth for a given time.

        :param location: location of the well to sonicate
        :type location: Union[Well, Tuple, Location]
        :param plunge_depth: depth (in mm) to plunge from the top of the plate.
        :type plunge_depth: float
        :param sonication_time: time (in sec) to sonicate for
        :type sonication_time: float
        :param power: sonicator power level ranging from 0.4 (default, min) through 1.0 (max).
        :type power: float
        :param pulse_duty_cycle: duty cycle of the sonicator pulse, defaults to 0.5
        :type pulse_duty_cycle: float, optional
        :param pulse_interval: interval between pulses, defaults to 1.0
        :type pulse_interval: float, optional
        :param verbose: whether or not to print out the time elapsed, defaults to False
        :type verbose: bool, optional
        :param autoclean: whether or not to perform the cleaning protocol after sonication, defaults to False
        :type autoclean: bool, optional
        :param args: if location is of type Tuple, the depth of the well needs to be specified
        :type args: tuple
        :raises ValueError: if the plunge depth is too deep

         Note: sonicator does not turn on below power level of 0.4.
        """

        # Check that plunger depth is compatible with labware dimensions
        if type(location) == Well:
            plate_height = location.top_
        elif type(location) == Location:
            plate_height = location[1].top_
        elif type(location) == Tuple:
            try:
                plate_height = args["depth"]
            except ValueError:
                print(
                    'If location is of type {}, parameter "depth" needs to be indicated'.format(
                        type(location)
                    )
                )

        plunge_height = plate_height - plunge_depth
        # Sanity check that we're not plunging too deep. Plunge depth is relative.

        self.plunge_depth = plunge_depth

        if plunge_height < 0:
            raise ValueError("Error: plunge depth is too deep.")

        if autoclean:
            assert self.cleaning is not None, "Error: cleaning protocol not set."

        x, y, z = Labware._getxyz(location)

        self._machine.safe_z_movement()
        self._machine.move_to(x=x, y=y)  # Position over the well at safe z height.
        self._machine.move_to(z=plunge_height, wait=True)
        print(f"Sonicating for {sonication_time} seconds!!")
        self._sonicate(
            sonication_time, power, pulse_duty_cycle, pulse_interval, verbose=verbose
        )
        print("done!")
        self._machine.safe_z_movement()

        if autoclean:
            self.perform_cleanining_protocol()

    def set_cleaning_protocol(
        self,
        wells: List[Well],
        power: List[float],
        time: List[float],
        plunge_depth: float = None,
    ):
        """Set the cleaning protocol for the sonicator.

        :param wells: list of wells to clean
        :type wells: List[Well]
        :param power: power level to clean at
        :type power: List[float]
        :param time: time to clean for
        :type time: List[float]
        :param plunge_depth: depth of the sonicator into the well, defaults to None
        :type plunge_depth: float, optional
        """
        assert (
            len(wells) == len(power) == len(time)
        ), "Error: wells, power, and time must be the same length."

        self.cleaning = {
            "Wells": wells,
            "Power": power,
            "Time": time,
            "Plunge Depth": plunge_depth,
        }
        print("Cleaning protocol set.")

    def perform_cleanining_protocol(self, plunge_depth: float = None):
        """Perform the cleaning protocol on the sonicator.

        :param plunge_depth: depth to plunge the sonicator into the well, defaults to None
        :type plunge_depth: float, optional

        """
        wells = self.cleaning["Wells"]
        power = self.cleaning["Power"]
        time = self.cleaning["Time"]
        depth = self.cleaning["Plunge Depth"]

        if depth != None:
            z_depth = depth
        elif plunge_depth != None:
            z_depth = plunge_depth
        elif self.plunge_depth != None and depth == None and plunge_depth == None:
            z_depth = self.plunge_depth
        else:
            assert plunge_depth != None, "Error: plunge depth must be specified."

        for well, power, time in zip(wells, power, time):
            self.sonicate_well(well, z_depth, time, power, 1, 1.0, verbose=True)

        self._machine.move_to(z=wells[-1].top(30))
        time.sleep(30)
        print("Cleaning protocol completed.")



import json
import os
import warnings
from typing import Tuple, Union

import numpy as np

from science_jubilee.labware.Labware import Labware, Location, Well
from science_jubilee.tools.Tool import (
    Tool,
    ToolConfigurationError,
    ToolStateError,
    requires_active_tool,
)


class Syringe(Tool):
    """A class representation of a syringe.

    :param Tool: The base tool class
    :type Tool: :class:`Tool`
    """

    def __init__(self, index, name, config):
        """Constructor method"""
        super().__init__(index, name)

        self.min_range = 0
        self.max_range = None
        self.mm_to_ml = None
        self.e_drive = "E"

        self.load_config(config)

    def load_config(self, config):
        """Loads the confirguration file for the syringe tool

        :param config: Name of the config file for your syringe. Expects the file to be in /tools/configs
        :type config: str
        """

        config_directory = os.path.join(os.path.dirname(__file__), "configs")
        config_path = os.path.join(config_directory, f"{config}.json")
        if not os.path.isfile(config_path):
            raise ToolConfigurationError(
                f"Error: Config file {config_path} does not exist!"
            )

        with open(config_path, "r") as f:
            config = json.load(f)
        self.min_range = config["min_range"]
        self.max_range = config["max_range"]
        self.mm_to_ml = config["mm_to_ml"]

        # Check that all information was provided
        if None in vars(self):
            raise ToolConfigurationError(
                "Error: Not enough information provided in configuration file."
            )

    def post_load(self):
        """Query the object model after loading the tool to find the extruder number of this syringe."""

        # To read the position of an extruder, we need to know which extruder number to look at
        # Query the object model to find this
        tool_info = json.loads(self._machine.gcode('M409 K"tools[]"'))["result"]
        for tool in tool_info:
            if tool["number"] == self.index:
                self.e_drive = (
                    f"E{tool['extruders'][0]}"  # Syringe tool has only 1 extruder
                )
            else:
                continue

    def check_bounds(self, pos):
        """Disallow commands outside of the syringe's configured range

        :param pos: The E position to check
        :type pos: float
        """

        if pos > self.max_range or pos < self.min_range:
            raise ToolStateError(f"Error: {pos} is out of bounds for the syringe!")

    @requires_active_tool
    def _aspirate(self, vol: float, s: int = 2000):
        """Aspirate a certain volume in milliliters. Used only to move the syringe; to aspirate from a particular well, see aspirate()

        :param vol: Volume to aspirate, in milliliters
        :type vol: float
        :param s: Speed at which to aspirate in mm/min, defaults to 2000
        :type s: int, optional
        """
        de = vol * -1 * self.mm_to_ml
        pos = self._machine.get_position()
        end_pos = float(pos[self.e_drive]) + de
        self.check_bounds(end_pos)
        self._machine.move(de=de, wait=True)

    @requires_active_tool
    def _dispense(self, vol, s: int = 2000):
        """Dispense a certain volume in milliliters. Used only to move the syringe; to dispense into a particular well, see dispense()

        :param vol: Volume to dispense, in milliliters
        :type vol: float
        :param s: Speed at which to dispense in mm/min, defaults to 2000
        :type s: int, optional
        """
        de = vol * self.mm_to_ml
        pos = self._machine.get_position()
        end_pos = float(pos[self.e_drive]) + de
        self.check_bounds(end_pos)
        self._machine.move(de=de, wait=True)

    @requires_active_tool
    def aspirate(
        self, vol: float, location: Union[Well, Tuple, Location], s: int = 2000
    ):
        """Aspirate a certain volume from a given well.

        :param vol: Volume to aspirate, in milliliters
        :type vol: float
        :param location: The location (e.g. a `Well` object) from where to aspirate the liquid from.
        :type location: Union[Well, Tuple, Location]
        :param s: Speed at which to aspirate in mm/min, defaults to 2000
        :type s: int, optional
        """
        x, y, z = Labware._getxyz(location)

        self._machine.safe_z_movement()
        self._machine.move_to(x=x, y=y)
        self._machine.move_to(z=z)
        self._aspirate(vol, s=s)

    @requires_active_tool
    def dispense(
        self, vol: float, location: Union[Well, Tuple, Location], s: int = 2000
    ):
        """Dispense a certain volume into a given well.

        :param vol:  Volume to dispense, in milliliters
        :type vol: float
        :param location: The location to dispense the liquid into.
        :type location: Union[Well, Tuple, Location]
        :param s: Speed at which to dispense in mm/min, defaults to 2000
        :type s: int, optional

        """
        x, y, z = Labware._getxyz(location)

        self._machine.safe_z_movement()
        self._machine.move_to(x=x, y=y)
        self._machine.move_to(z=z)
        self._dispense(vol, s=s)

    @requires_active_tool
    def transfer(
        self,
        vol: float,
        s: int = 2000,
        source: Well = None,
        destination: Well = None,
        mix_before: tuple = None,
        mix_after: tuple = None,
    ):
        """Transfer liquid from source well(s) to a set of destination well(s). Accommodates one-to-one, one-to-many, many-to-one, and uneven transfers.

        :param vol: Volume to transfer in milliliters
        :type vol: float
        :param s: Speed at which to aspirate and dispense in mm/min, defaults to 2000
        :type s: int, optional
        :param source: A source well or set of source wells, defaults to None
        :type source: Well, optional
        :param destination: A destination well or set of destination wells, defaults to None
        :type destination: Well, optional
        :param mix_before: Mix the source well before transfering, defaults to None
        :type mix_before: tuple, optional
        :param mix_after: Mix the destination well after transfering, defaults to None
        :type mix_after: tuple, optional
        """
        if type(source) != list:
            source = [source]
        if type(destination) != list:
            destination = [destination]

        # Assemble tuples of (source, destination)
        num_source_wells = len(source)
        num_destination_wells = len(destination)
        if num_source_wells == num_destination_wells:  # n to n transfers
            pass
        elif (
            num_source_wells == 1 and num_destination_wells > 1
        ):  # one to many transfers
            source = list(np.repeat(source, num_destination_wells))
        elif (
            num_source_wells > 1 and num_destination_wells == 1
        ):  # many to one transfers
            destination = list(np.repeat(destination, num_source_wells))
        elif num_source_wells > 1 and num_destination_wells > 1:  # uneven transfers
            # for uneven transfers, find least common multiple to pair off wells
            # raise a warning, as this might be a mistake
            # this mimics OT-2 behavior
            least_common_multiple = np.lcm(num_source_wells, num_destination_wells)
            source_repeat = least_common_multiple / num_source_wells
            destination_repeat = least_common_multiple / num_destination_wells
            source = list(np.repeat(source, source_repeat))
            destination = list(np.repeat(destination, destination_repeat))
            warnings.warn("Warning: Uneven source & destination wells specified.")

        source_destination_pairs = list(zip(source, destination))
        for source_well, destination_well in source_destination_pairs:
            # TODO: Large volume transfers which exceed tool capacity should be split up into several transfers
            xs, ys, zs = Labware._getxyz(source_well)
            xd, yd, zd = Labware._getxyz(destination_well)

            self._machine.safe_z_movement()
            self._machine.move_to(x=xs, y=ys)
            self._machine.move_to(z=zs + 5)
            self.current_well = source_well
            self._aspirate(vol, s=s)

            #             if mix_before:
            #                 self.mix(mix_before[0], mix_before[1])
            #             else:
            #                 pass

            self._machine.safe_z_movement()
            self._machine.move_to(x=xd, y=yd)
            self._machine.move_to(z=zd + 5)
            self.current_well = destination_well
            self._dispense(vol, s=s)


#             if mix_after:
#                 self.mix(mix_after[0], mix_after[1])
#             else:
#                 pass



import os

# from picamera2 import Picamera2
import time

from flask import Flask, send_file, send_from_directory

scratch_fp = "./"


def create_app():

    app = Flask(__name__)
    from picamera2 import Picamera2

    camera = Picamera2()

    @app.route("/image", methods=("GET", "POST"))
    def image():
        """
        Take an image and return it over the function
        """
        time.sleep(2)
        camera.start_and_capture_file(scratch_fp + "capture.jpg")

        return send_file("../capture.jpg")  # , as_attachement=True)

    return app



import json
import os
import string
from dataclasses import dataclass
from itertools import chain
from math import acos, cos, sin, sqrt
from typing import Dict, Iterable, List, NamedTuple, Tuple, Union

import numpy as np


@dataclass
class Well:
    """A class representing a well of a labware.

    Each Well is associated with a specific name, depth, total liquid volume, shape, diameter,
    x, y, and z dimension, y-dimension, as well as its coordinates and any applied offset

    :return: A :class:`Well` object with various information about the geometry of the well and its position in the labware
    :rtype: :class:`Well`
    """

    name: str
    depth: float
    totalLiquidVolume: float
    shape: str
    diameter: float = None
    xDimension: float = None
    yDimension: float = None
    x: float
    y: float
    z: float
    offset: Tuple[float] = None
    slot: int = None
    has_tip: bool = False
    clean_tip: bool = False
    labware_name: str = None

    @property
    def x(self):
        """Offsets the x-position of the each well with respect to the deck-slot coordinates

        :return: The x-coordinate of the well
        :rtype: float
        """
        return self._x

    @x.setter
    def x(self, new_x):
        """Setter for the offsetted x-position of each well with respect to the deck-slot coordinates

        :param new_x: the new y-coordinate of the well
        :type new_x: float
        """
        self._x = new_x

    @property
    def y(self):
        """Offsets the y-position of the each well with respect to the deck-slot coordinates

        :return: The y-coordinate of the well
        :rtype: float
        """
        return self._y

    @y.setter
    def y(self, new_y):
        """Setter for the offsetted y-position of each well with respect to the deck-slot coordinates

        :param new_y: The new y-coordinate of the well
        :type new_y: float
        """

        self._y = new_y

    @property
    def z(self):
        """Offsets the z-position of each well with respect to the deck-slot coordinates

        :return: The z-coordinate of the well
        :rtype: float
        """
        return self._z

    @z.setter
    def z(self, new_z):
        """Setter for the offsetted z-position of each well with respect to the deck-slot coordinates

        :param new_z: The new z-coordinate of the well
        :type new_z: flaot
        """
        self._z = new_z

    def apply_offset(self, offset: Tuple[float]):
        """Allows the user to offset the coordinates of the well with respect to the deck-slot coordinates

        :param offset: A tuple of floats with the new offset of the well
        :type offset: Tuple[float]
        """
        self._x = self.x + offset[0]
        self._y = self.y + offset[1]

        if len(offset) == 3:
            self._z = self.z + offset[2]

        self.offset = offset

    @property
    def top_(self):
        """Defines the top-most point of the well

        :return: The z-coordinate of the top of the well
        :rtype: float
        """
        return self.z + self.depth

    @property
    def bottom_(self):
        """Defines the bottom-most point of the well

        :return: The z-coordinate of the bottom of the well
        :rtype: float
        """
        return self.z

    def bottom(self, z: float, check=False):
        """Allows the user to dinamically indicate a new Z location relative to the
        bottom of the well.

        :param z: the distance in mm to offset the coordinates from the bottom of the well. Should be +
        :type z: float
        :param check: the 'z' parameters can either be + or -. If negative, an assert error is raised to
            avoid collision with the labware. However, there might be instances of custom labware where the bottom of the well
            is purposely set as higher during the generation of its config .json file., defaults to False
        :type check: bool, optional
        :return: A :class:`Location` which contains information about the new coordinates generated and the
            :class:`Well` object
        :rtype: :class:`Location`
        """
        from_bottom_z = self.bottom_ + z
        if check:
            pass
        else:
            assert z >= 0, (
                "Indicated location is lower than the bottom of the labware and"
                " could result in crash. Input a positive 'z' value "
            )

        coord = (self.x, self.y, from_bottom_z)

        return Location(coord, self)

    def top(self, z: float):
        """Allows the user to dinamically indicate a new Z location relative to the
        top of the well.

        :param z: the distance in mm to offset the coordinates from the top of the well.Can be either + or -
        :type z: float
        :return: A :class:`Location` which contains information about the new coordinates generated and the
            :class:`Well` object.
        :rtype: :class:`Location`
        """
        from_top_z = self.top_ + z
        assert (
            from_top_z > self.bottom_
        ), "Indicated location is lower than the bottom of the labware."
        coord = (self.x, self.y, from_top_z)

        return Location(coord, self)

    def __repr__(self):
        """Displayed representation of a :class:`Well` object indicating its name and its coordinates

        :return: A string representation of the name and coordinates of a well
        :rtype: str
        """
        if self.slot != None:
            message = f"Well {self.name} form {self.labware_name} on slot {self.slot}"
        else:
            message = f"Well {self.name} at coordinates {self.x, self.y, self.z}"
        return message

    def set_has_tip(self, value: bool):
        """Set the value of the `has_tip` attribute.

        :param value: The new value for the `has_tip` attribute
        :type value: bool
        """
        self.has_tip = value

    def set_clean_tip(self, value: bool):
        """Returns the value of the `clean_tip` attribute.

        :param value: The new value for the `clean_tip` attribute
        :type value: bool
        """
        self.clean_tip = value


@dataclass(repr=False)
class WellSet:
    """A class defining a set of wells expressed as a dictionary in which each keys is the
    the :attribute:`Well.name` object and the value is the :class:`Well` object itself.
    """

    wells: Dict[str, Well]

    def __repr__(self):
        """Displays the wellset as a :list: of wells and the deck-slot nunmber

        :return: A :list: of :class:`Well` objects diplayed by their :attribute:`Well.name`
        :rtype: :class:`Row`
        """
        return str(f"{list(self.wells.keys())}")

    def __getitem__(self, id_: Union[str, int]):
        """Allows the user to select a :class:`Well` object by either their :attribute:`Well.name` or
            their index in a :list:

        :param id_: The :attribute:`Well.name` or index representing a :class:`Well` in the labware
        :type id_: Union[str, int]
        :return: The :class:`Well` object
        :rtype: :class:`Well`
        """
        try:
            if isinstance(id_, slice):
                well_list = []
                start = id_.start
                stop = id_.stop
                if id_.step is not None:
                    step = id_.step
                else:
                    step = 1
                for sub_id in range(start, stop, step):
                    well_list.append(self.wells[sub_id])
                return well_list
            else:
                return self.wells[id_]
        except KeyError:
            return list(self.wells.values())[id_]


@dataclass(repr=False)
class Row(WellSet):
    """A class representing a row of a labware, for example 'A', 'B', etc

    :param WellSet: A dictionary of :class:`Well` objects in which each keys is the the :attribute:`Well.name` object and the value is the :class:`Well` object itself.
    :type WellSet: Dict[str, Well]
    """

    identifier: str


@dataclass(repr=False)
class Column(WellSet):
    """A class representing a column of a labware, for example 1, 2, etc.

    :param WellSet: A dictionary of :class:`Well` objects in which each keys is the the :attribute:`Well.name` object and the value is the :class:`Well` object itself.
    :type WellSet: Dict[str, Well]
    """

    identifier: int


class Labware(WellSet):
    """A class representing a basic laboratory labware made up of a set of wells/pipette tips.

    :param labware_filename: The name of the config `.json`
    :type labware_filename: str
    :param offset: Coordinates to use to offset all the wells in a labware for easier handling of coordinates.
        For example this is called by the :method:`Deck.load_labware` when assignign a labware to a deck slot, defaults to None
    :type offset: Tuple[float], optional
    :param order: Option to order the wells of a labware either by `row` or `columns`, defaults to 'rows'
    :type order: str, optional
    :param path: Path to the folder containing the configuration `.json` files for the labware,
            defaults to the 'labware_definition/' in the science_jubilee/labware directory.
    :type path: str, optional
    """

    def __init__(
        self,
        labware_filename: str,
        offset: Tuple[float] = None,
        order: str = "rows",
        path: str = os.path.join(os.path.dirname(__file__), "labware_definition"),
    ):
        """Initializes a :class:`Labware` object by loading its configuration file and creating a dictionary of :class:`Well` objects.

        :param labware_filename: The name of the config `.json`
        :type labware_filename: str
        :param offset: Coordinates to use to offset all the wells in a labware for easier handling of coordinates.
            For example this is called by the :method:`Deck.load_labware` when assignign a labware to a deck slot, defaults to None
        :type offset: Tuple[float], optional
        :param order: Option to order the wells of a labware either by `row` or `columns`, defaults to 'rows'
        :type order: str, optional
        :param path: Path to the folder containing the configuration `.json` files for the labware,
                defaults to the 'labware_definition/' in the science_jubilee/labware directory.
        :type path: str, optional
        """
        # load in the labware configuration file
        if labware_filename[-4:] != "json":
            labware_filename = labware_filename + ".json"

        config_path = os.path.join(path, f"{labware_filename}")

        with open(config_path, "r") as f:
            # this will be the raw .json file data and fields should not be modified directly
            # current exceptions is 'manual_offset' field to allow to save custom data for easier handling
            # of recurrent slot-labware combinations
            self.data = json.load(f)

        self.config_path = config_path
        self.wells_data = self.data.get("wells", {})
        self.row_data, self.column_data, self.wells = self._create_rows_and_columns()

        order_options = [
            "rows",
            "row",
            "Rows",
            "Row",
            "R",
            "cols",
            "col",
            "C",
            "columns",
            "Columns",
        ]
        assert order in order_options, "Order must be one of {}".format(order_options)
        self.withWellOrder(order)
        self.offset = offset
        self.slot = None

        # check to see if a manual offset was saved for this labware in a specific slot
        if "manual_offset" in self.data:
            self.manualOffset = self.data["manual_offset"]
        else:
            # otherwise initialize manual_offset instance variable
            self.manualOffset = {}

    def __repr__(self):
        """Displayed representation of a :class:`Labware` object indicating the type of labware and
        its name. Additionally, it will show the :attribute:`Deck.slots` number if the labware has been
        already assigned to it.
        """
        display = (
            self.metadata()["displayCategory"] + ": " + self.parameters()["loadName"]
        )
        if self.slot is not None:
            display = display + " " + f" on {self.slot}"
        return display

    def _create_rows_and_columns(self):
        """Creates a dictionary of :class:`Row` and :class:`Column` and :class:`Well` objects from the data in the config `.json` file.

        :return: A dictionary of :class:`Row` and :class:`Column` and :class:`Well` objects
        :rtype: :class:`Row`, :class:`Column`, :class:`Well`
        """
        rows = {}
        columns = {}
        wells = {}

        for row_order, column_data in enumerate(self.ordering):
            # Assumes the first char is the row identifier, e.g., "A" in "A1"
            row_id = column_data[0][0]
            # Extracts column number, e.g., "1" in "A1"
            col_ids = [int(well[1:]) for well in column_data]

            if row_id not in rows:
                rows[row_id] = {}

            for col_order, well_id in enumerate(column_data):
                well = Well(name=well_id, **self.wells_data[well_id])
                rows[row_id][well_id] = well

                if col_order + 1 not in columns:  # +1 since indexing starts at 0
                    columns[col_order + 1] = {}

                columns[col_order + 1][well_id] = well
                wells[well_id] = well

        # add tip tracking to the wells
        if self.is_tip_rack:
            for well in wells.values():
                well.has_tip = True
                well.clean_tip = True

        # add labware name to each Well object
        for well in wells.values():
            well.labware_name = self.display_name

        # Convert dictionary data to Row and Column classes
        _rows = {k: Row(identifier=k, wells=v) for k, v in rows.items()}
        _columns = {k: Column(identifier=k, wells=v) for k, v in columns.items()}

        return _rows, _columns, wells

    def get_row(self, row_id: str) -> Row:
        """Fucntions to fetch the :class:`Well.name` of the indicated row.

        :param row_id: The name of a row of the labware, usually indicated by a capital letter (e.g., A, B, etc.)
        :type row_id: str
        :return: A list of :class:`Well` objects diplayed by their :attribute:`Well.name`
        :rtype: :class:`Row`
        """
        return self.row_data.get(row_id)

    def get_column(self, col_id: int) -> Column:
        """Fucntions to fetch the :class:`Well.name` of the indicated column.

        :param col_id: The name of a column of the labware, usually indicated by an integer number (e.g., 1, 2, etc.)
        :type col_id: str
        :return: A list of :class:`Well` objects diplayed by their :attribute:`Well.name`
        :rtype: :class:`Column`
        """
        return self.column_data.get(col_id)

    @property
    def shape(self):
        """Returns the shape of the labware as a tuple of (rows, columns)

        :return: A tuple of (rows, columns)
        :rtype: Tuple[int, int]
        """
        return (len(self.row_data), len(self.column_data))

    @property
    def ordering(self) -> List[List[str]]:
        """Returns the ordering of the wells in the labware as a list of lists. Each list represents a row of the labware.

        :return: A list of lists of :class:`Well.name` objects
        :rtype: List[List[str]]
        """

        return np.array(self.data["ordering"]).T

    @property
    def brand(self) -> dict:
        """Returns the brand of the labware as a strin

        :return: A string with the brand of the labware
        :rtype: str
        """
        return self.data.get("brand", {})["brand"]

    def metadata(self) -> dict:
        """Returns the metadata of the labware as a dictionary

        The metadata of a labware will generally contain the display name, the type of labware, and the units of volume.
        These can also be found as attributes of the :class:`Labware` object.

        :return: A dictionary with the metadata of the labware
        :rtype: dict
        """
        return self.data.get("metadata", {})

    @property
    def display_name(self):
        """Returns the display name of the labware as a string

        :return: A string with the display name of the labware
        :rtype: str
        """
        return self.metadata()["displayName"]

    @property
    def labware_type(self):
        """Returns the type of labware as a string

        The type fo labware will generally either be a tiprack, wellplate, reservoir, etc.

        :return: A string with the type of labware
        :rtype: str
        """
        return self.metadata()["displayCategory"]

    @property
    def volume_units(self):
        """Returns the units of volume of the labware as a string

        The volume units will be either uL or mL.

        :return: A string with the units of volume of the labware
        :rtype: str
        """
        return self.metadata()["displayVolumeUnits"]

    @property
    def dimensions(self) -> dict:
        """Returns the dimensions of the labware as a dictionary

        :return: A dictionary with the x,y, and z dimensions of the labware
        :rtype: dict
        """
        return self.data.get("dimensions", {})

    def parameters(self) -> dict:
        """Returns the parameters describing certain features of the labware as a dictionary

        The parameters genereally include whether the shape of the labware is regular or irregular, if it is a tiprack,
        and other Opentrons specific parameters as we are using their 'Custom Labware Page' to generate the .json config files.

        :return: A dictionary with the parameters of the labware
        :rtype: dict
        """
        return self.data.get("parameters", {})

    @property
    def is_tip_rack(self):
        """Returns a boolean indicating if the labware is a tiprack

        :return: True if the labware is a tiprack, False otherwise
        :rtype: bool
        """
        return self.parameters()["isTiprack"]

    @property
    def load_name(self):
        """Returns the name of the labware as a string

        :return: A string with the name of the labware
        :rtype: str
        """
        return self.parameters()["loadName"]

    @property
    def tip_length(self):
        """Returns the length of the tip of the labware as a float if the labware is a tiprack, otherwise returns None

        :return: A float with the length of the tip of the labware or None otherwise
        :rtype: float
        """
        try:
            return self.parameters()["tipLength"]
        except:
            pass

    @property
    def tip_overlap(self):
        """Returns the overlap of the tip of the labware as a float if the labware is a tiprack, otherwise returns None

        :return: A float with the overlap of the tip of the labware or None otherwise
        :rtype: float
        """
        try:
            return self.parameters()["tipOverlap"]
        except:
            pass

    @property
    def offset(self):
        """Returns the offset of the labware as a tuple of floats

        :return: A tuple of floats with the offset of the labware
        :rtype: Tuple[float]
        """
        return self._offset

    @offset.setter
    def offset(self, new_offset):
        """Sets the offset of the labware to the indicated values and updates the offset of each well in the labware

        :param new_offset: A tuple of floats with the new offset of the labware
        :type new_offset: Tuple[float]
        """
        self._offset = new_offset
        if new_offset is not None:
            for w in self:
                w.apply_offset(new_offset)

    def add_slot(self, slot_):
        """Add name of deck slot after labware has been loaded

        :param slot_: The name of the deck slot
        :type slot_: str
        """
        self.slot = slot_
        for w in self:
            w.slot = slot_

    def withWellOrder(self, order) -> list:
        """Reorders the wells by rows or by columns. Automatically updates the :attribute:`Labware.wells`

        :param order: The order in which to reorder the wells. Can be either 'rows' or 'columns'
        :type order: str
        :return: A list of :class:`Well` objects diplayed by their :attribute:`Well.name`
        :rtype: list
        """
        ordered_wells = {}
        if order in ["rows", "row", "Rows", "Row", "R"]:
            for well in list(chain(*self.row_data.values())):
                ordered_wells[well.name] = well
        elif order in ["cols", "col", "C", "columns", "Columns"]:
            for well in list(chain(*self.column_data.values())):
                ordered_wells[well.name] = well
        else:
            print("Order needs to be either rows or columns")

        self.wells = ordered_wells

    # @staticmethod
    def _translate_point(
        self,
        well: Well,
        theta: float,
        x_space: float,
        y_space: float,
        upper_left: Tuple[float],
    ):
        """
        Helper function to translate the coordinates of a well by a given angle theta.

        :param well: A :class:`Well` object
        :type well: :class:`Well`
        :param theta: The angle by which to translate the coordinates of the well
        :type theta: float

        :return: The new x and y coordinates of the well
        :rtype: float, float
        """
        x_nom, y_nom = self._nominal_coordinates(well, x_space, y_space)

        x_translated = upper_left[0] + x_nom * cos(theta) - y_nom * sin(theta)
        y_translated = upper_left[1] - (x_nom * sin(theta) + y_nom * cos(theta))

        return x_translated, y_translated

    @staticmethod
    def _nominal_coordinates(well: Well, x_space: float, y_space: float):
        """
        Helper function to calculate the nominal coordinates of a well in a labware
        based on its row and column index.
        """
        col_index = int(well.name[1:]) - 1
        row_index = list(string.ascii_uppercase).index(well.name[0])

        x_nominal = col_index * x_space
        y_nominal = row_index * y_space

        return x_nominal, y_nominal

    def manual_offset(self, corner_wells: List[Tuple[float]], save: bool = False):
        """Allows the user to manually offset the coordinates of the labware based on three corner wells.

        Adapted from `https://github.com/machineagency/sonication_station` labware calibration procedure.

        :param offset: A list containing tuples of floats
        :type offset: Tuple[float]
        :param save: Option to save the manual offset to the original config `.json` file, defaults to False
        :type save: bool, optional

        :return: An updated :class:`Labware` object with the new coordinates of the wells
        :rtype: :class:`Labware`
        """
        assert (
            self.slot is not None
        ), "Labware has not been assigned to a slot yet. Use the 'add_slot' method to assign a slot"

        assert len(corner_wells) == 3, "Three points needed to apply manual offset"
        assert all(
            [len(o) == 2 for o in corner_wells]
        ), "Each point should have three coordinates (x,y)"

        # Get the coordinates of the three corner wells (e.g., A1, A12, H12)
        upper_left = corner_wells[0]
        upper_right = corner_wells[1]
        bottom_right = corner_wells[2]

        # Get the coordinates of the three corner wells
        # calculate total spacing between wells in each row (width) and column (height)
        plate_width = sqrt(
            (upper_right[0] - upper_left[0]) ** 2
            + (upper_right[1] - upper_left[1]) ** 2
        )
        plate_height = sqrt(
            (bottom_right[0] - upper_right[0]) ** 2
            + (bottom_right[1] - upper_right[1]) ** 2
        )

        # Assume evenly spaced wells, but possible to have different spacing in rows and columns
        x_space = plate_width / (len(self.column_data) - 1)
        y_space = plate_height / (len(self.row_data) - 1)

        # Define and average the offset angles for the plate
        theta1 = acos((upper_right[1] - bottom_right[1]) / plate_height)
        theta2 = acos((upper_right[0] - upper_left[0]) / plate_width)
        theta = (theta1 + theta2) / 2.0
        # apply offset to all wells in the labware object

        for well in self:
            new_x, new_y = self._translate_point(
                well, theta, x_space, y_space, upper_left
            )
            well.x = new_x
            well.y = new_y
        print(f'New manual offset applied to {self.parameters()["loadName"]}')

        if save:
            if str(self.slot) in self.manualOffset.keys():
                k = input(
                    "Are you sure you want to overwrite the manual offset for this labware? Press 'y' key to continue"
                )
                if k == "y":
                    self.manualOffset[str(self.slot)] = corner_wells
                    with open(self.config_path, "w") as f:
                        self.data["manual_offset"] = {str(self.slot): corner_wells}
                        json.dump(self.data, f)
                    print("Manual offset saved")
                else:
                    print("Manual offset applied, but not saved")
            else:
                self.manualOffset[str(self.slot)] = corner_wells
                with open(self.config_path, "w") as f:
                    self.data["manual_offset"] = {str(self.slot): corner_wells}
                    f.seek(0)
                    json.dump(self.data, f, indent=4)
                print("Manual offset saved")
        else:
            self.manualOffset[str(self.slot)] = corner_wells

    def load_manualOffset(self, apply: bool = True):
        """Loads the manual offset of a labware from its config `.json` file for a specific slot

        :param apply: Option to apply the manual offset to the labware or return values, defaults to False
        :type apply: bool, optional

        :return: A list of tuples containing the manual offset of the labware
        :rtype: List[Tuple[float]]
        """
        assert (
            self.slot is not None
        ), "Labware has not been assigned to a slot yet. Use the 'add_slot' method to assign a slot"
        if self.manualOffset[str(self.slot)]:
            if apply:
                self.manual_offset(self.manualOffset[str(self.slot)])
                return
            else:
                return self.manualOffset[str(self.slot)]
        else:
            return self.data["manual_offset"][self.slot]

    @staticmethod
    def _getxyz(location: Union[Well, Tuple, "Location"]):
        """Helper function to extract the x, y, z coordinates of a location object.

        :param location: The location object to extract the coordinates from. This can either be a
            :class:`Well`, a :tuple: of x, y, z coordinates, or a :class:`Location` object
        :type location: Union[Well, Tuple, Location]
        :raises ValueError: If the location is not a :class:`Well`, a :class:`tuple`, or a :class:`Location` object
        :return: The x, y, z coordinates of the location
        :rtype: float, float, float
        """
        if type(location) == Well:
            x, y, z = location.x, location.y, location.z
        elif type(location) == tuple:
            x, y, z = location
        elif type(location) == Location:
            x, y, z = location._point
        else:
            raise ValueError("Location should be of type Well or Tuple")

        return x, y, z


## Adapted from Opentrons API  opentrons.types##
class Point(NamedTuple):
    """A point in the Jubilee 3D coordinate system.

    :param NamedTuple: A list-like container with a fixed number of elements
    :type NamedTuple: :class:`NamedTuple`
    :return: A tuple of coordinates (x,y,z)
    :rtype: :class:`Point`
    """

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def add(self, other):
        """Adds the coordinates of two points

        :param other: A tuple of coordinates (x,y,z)
        :type other: :class:`Point`
        :return: A new :class:`Point` object
        :rtype: :class:`Point`
        """
        if not isinstance(other, Point):
            return NotImplemented
        return Point(self.x + other.x, self.y + other.y, self.z + other.z)

    def substract(self, other):
        """Substracts the coordinates of two points

        :param other: A tuple of coordinates (x,y,z)
        :type other: :class:`Point`
        :return: A new :class:`Point` object
        :rtype: :class:`Point`
        """
        if not isinstance(other, Point):
            return NotImplemented
        return Point(self.x - other.x, self.y - other.y, self.z - other.z)

    def multiply(self, other: Union[int, float]):
        """Multiplies the coordinates of a point by a scalar

        :param other: A scalar to multiply the coordinates of a point
        :type other: Union[int, float]
        :return: A new :class:`Point` object scaled by the value indicated as the function parameter
        :rtype: :class:`Point`
        """

        if not isinstance(other, (float, int)):
            return NotImplemented
        return Point(self.x * other, self.y * other, self.z * other)

    def absolute(self):
        """Returns the absolute value of the coordinates of a point.

        :return: The absolute values of a :class:`Point` object
        :rtype: :class:`Point`
        """
        return Point(abs(self.x), abs(self.y), abs(self.z))

    def __repr__(self) -> str:
        """Returns a string representation of the coordinates of a point.

        :return: A string representation of the coordinates of a point
        :rtype: str
        """

        display = "x:{}, y: {}, z:{}".format(self.x, self.y, self.z)
        return display


class Location:
    """A location to target as a motion.

    The location contains a :class:`Point` and possibly an associated
    :class:`Labware` or :class:`Well` instance.
    """

    def __init__(self, point: Point, labware: Union[Well, Labware]):

        self._point = point
        self._labware = labware

    @property
    def point(self) -> Point:
        """The coordinates (x,y,z) of a Well or a Labware

        :return: A tuple of coordinates (x,y,z)
        :rtype: :class:`Point`
        """
        return self._point

    @property
    def labware(self):
        """The :class:`Well` object associated with the coordinates (x,y,z)

        :return: A :class:`Well` object
        :rtype: :class:`Well`
        """
        return self._labware

    def __iter__(self) -> Iterable[Union[Point, Well, Labware]]:
        """Iterable interface to support unpacking of :class:`Location` objects.

        :return: An interable of :class:`Location` objects
        :rtype: Iterable[Union[Point, Well, Labware]]
        """
        return iter((self._point, self._labware))

    def __eq__(self, other: object) -> bool:
        """Comparison between two :class:`Location` objects.

        :param other: A :class:`Location` object
        :type other: :class:`Location`
        :return: True if the two :class:`Location` objects are equal, False otherwise
        :rtype: bool
        """
        return (
            isinstance(other, Location)
            and other._point == self._point
            and other._labware == self._labware
        )

    def __repr__(self) -> str:
        """Returns a string representation of the :class:`Location` object.

        :return: A string representation of the :class:`Location` object
        :rtype: str
        """
        return f"Location(point={repr(self._point)}, labware={self._labware})"



import platform
import time

import cv2
import matplotlib
import numpy as np

matplotlib.use("TkAgg")
from typing import Tuple

import yaml
from matplotlib import pyplot as plt

from science_jubilee.labware.Labware import Well
from science_jubilee.tools.Tool import Tool, requires_active_tool

if platform.system() == "Linux":
    import picamera  # Note that this can only be installed on raspbery pi.


class Camera(Tool):
    """A class representation of a Raspberry Pi camera.

    :param Tool: The base tool class
    :type Tool: class:`Tool`
    """

    def __init__(self, index, name):
        """Constructor method"""
        super().__init__(index, name)
        self._camera_matrix = None
        self._dist_matrix = None
        # self.load_coefficients(
        #     "/home/pi/POSE/science-jubilee/science-jubilee/tools/configs/calibration_checkerboard.yml"
        # )

    def load_coefficients(self, path):
        """Loads camera matrix and distortion coefficients.

        :param path: Path to your camera calibration file
        :type path: str
        :return: A list containing your camera matrix (index 0) and distortion matrix (index 1)
        :rtype: list
        """ """"""
        with open(path, "r") as file:
            config = yaml.safe_load(file)

        k_data = config["K"]["data"]
        camera_matrix = np.array(
            [k_data[i : i + 3] for i in range(0, len(k_data), 3)], dtype=object
        )

        d_data = config["D"]["data"]
        dist_matrix = np.array(
            [d_data[i : i + 3] for i in range(0, len(d_data), 3)], dtype=object
        )
        self._camera_matrix = camera_matrix
        self._dist_matrix = dist_matrix

        return [camera_matrix, dist_matrix]

    def get_camera_indices(self):
        """Returns valid camera indices for use with OpenCV

        :return: A list of valid camera indices
        :rtype: list
        """
        index = 0
        arr = []
        i = 4
        while i > 0:
            try:
                cap = cv2.VideoCapture(index)
                if cap.read()[0]:
                    arr.append(index)
                    cap.release()
            except:
                print("exception")
            index += 1
            i -= 1
        return arr

    @requires_active_tool
    def get_frame(self, resolution=[1200, 1200], uvc=False):
        """Take a picture and return the image. Compensates for lens distortion using camera calibration file.

        :param resolution: Camera resolution, defaults to [1200, 1200]
        :type resolution: list, optional
        :param uvc: True if the camera is a USB video class (UVC) camera for programmatically setting focus, defaults to False
        :type uvc: bool, optional
        :return: The captured frame
        :rtype: ndarray
        """
        with picamera.PiCamera() as camera:
            camera.resolution = (1200, 1200)
            camera.framerate = 24
            time.sleep(5)
            output = np.empty((resolution[1], resolution[0], 3), dtype=np.uint8)
            camera.capture(output, "rgb", use_video_port=True)
            tpose = np.transpose(output, axes=(1, 0, 2))
            # undistorted = cv2.undistort(
            #     tpose, self._camera_matrix, self._dist_matrix, None, self._camera_matrix
            # )
            return tpose

    def show_frame(self, frame, grid=False, save=False, save_path="fig.png"):
        """Show a captured frame using matplotlib.

        :param frame: The captured frame to show
        :type frame: ndarray
        :param grid: Show grid lines, defaults to False
        :type grid: bool, optional
        :param save: Save to file, defaults to False
        :type save: bool, optional
        :param save_path: File path to save image, defaults to "fig.png"
        :type save_path: str, optional
        """
        plt.imshow(frame)
        plt.title("frame capture")
        if grid:
            plt.grid()  # add a grid
            h, w, z = frame.shape
            plt.plot(
                [w / 2], [h / 2], marker="o"
            )  # put a marker in the center of the image
        if save:
            plt.axis("off")
            plt.savefig(f"{save_path}")
        plt.show()

    def get_show_frame(self):
        """Get and show a frame."""
        self.show_frame(self.get_frame())

    @requires_active_tool
    def video_stream(self, camera_index=0):
        """Start a video stream from the camera.

        :param camera_index: The camera index, defaults to 0
        :type camera_index: int, optional
        """
        cap = cv2.VideoCapture(
            camera_index
        )  # Note that the index corresponding to your camera may not be zero but this is the most common default

        # draw a circle in the center of the frame
        center = None
        while center is None:
            # the first frame grab is sometimes empty
            ret, frame = cap.read()
            h, w = frame.shape[0:2]
            center = (int(w / 2), int(h / 2))

        while True:
            ret, frame = cap.read()
            target = cv2.circle(frame, center, 5, (0, 255, 0), -1)
            cv2.imshow("Input", frame)
            c = cv2.waitKey(1)
            if (
                c == 27
            ):  # 27 is the built in code for ESC so press escape to close the window.
                break

        cap.release()
        cv2.destroyAllWindows()

    @requires_active_tool
    def image_wells(self, resolution=[1200, 1200], uvc=False, wells: Well = None):
        """Move to a number of wells to take and show images.

        :param resolution: Camera resolution, defaults to [1200, 1200]
        :type resolution: list, optional
        :param uvc: True if the camera is a USB video class (UVC) camera for programmatically setting focus, defaults to False
        :type uvc: bool, optional
        :param wells: A list of wells to image, defaults to None
        :type wells: :class:`Well`, optional
        """
        # TODO: different functions for saving many images, showing images, or getting frames for analysis?
        if type(wells) != list:
            wells = [wells]

        for well in wells:
            x, y, z_bottom = self._get_xyz(well=well)
            self._machine.safe_z_movement()
            self._machine.move_to(x=x, y=y)
            self._machine.move_to(z=30)  # focus height; read in from config
            time.sleep(1)  # ToDo: Better way to sync gcode movements & images
            f = self.get_frame()
            self.show_frame(f)

    @requires_active_tool
    def get_well_image(self, resolution=[1200, 1200], uvc=False, well: Well = None):
        """Move to a single well to take a picture and return the frame.

        :param resolution: Camera resolution, defaults to [1200, 1200]
        :type resolution: list, optional
        :param uvc: True if the camera is a USB video class (UVC) camera for programmatically setting focus, defaults to False
        :type uvc: bool, optional
        :param well: The well to image, defaults to None
        :type well: :class:`Well`, optional
        :return: The captured frame
        :rtype: ndarray
        """
        x, y, z_bottom = self._get_xyz(well=well)
        self._machine.safe_z_movement()
        self._machine.move_to(x=x, y=y)
        self._machine.move_to(z=30)  # focus height; read in from config
        time.sleep(1)  # ToDo: Better way to sync gcode movements & images
        f = self.get_frame()
        return f

    @staticmethod
    def _get_xyz(well: Well = None, location: Tuple[float] = None):
        """Get the (x,y,z) position of a well.

        :param well: The well to fetch position of, defaults to None
        :type well: :class:`Well`, optional
        :param location: Directly specify an (x,y,z) location, defaults to None
        :type location: Tuple[float], optional
        :raises ValueError: Must specify either a well or a location
        :return: The well location
        :rtype: Tuple[float, float, float]
        """
        if well is not None and location is not None:
            raise ValueError("Specify only one of Well or x,y,z location")
        elif well is not None:
            x, y, z = well.x, well.y, well.z
        else:
            x, y, z = location
        return x, y, z

    @staticmethod
    def _get_top_bottom(well: Well = None):
        """Get the top and bottom heights of a well.

        :param well: The well to fetch position of, defaults to None
        :type well: Well, optional
        :return: The z-height of the top and bottom of the well
        :rtype: Tuple[float, float]
        """
        top = well.top
        bottom = well.bottom
        return top, bottom



import json
import logging
import os
from typing import Tuple, Union

from science_jubilee.labware.Labware import Labware, Location, Well
from science_jubilee.tools.Tool import Tool


class PumpDispenser(Tool):
    """
    Class to manage dispenser tool, for example for a color mixing demo

    :param index: The tool index of the dispenser head on the machine
    :type indes: int
    :param name: The tool name
    :type name: str
    :param pump_group: science_jubilee.tools.PeristalticPump.PeristalticPumps object to assign to dispenser head
    :type pump_group: science_jubilee.tools.PeristalticPump.PeristalticPumps
    :param dispense_tip_offset: Offsets from tool reference point for every dispense tip on the tool
    :type dispense_tip_offset: list
    :param line_volume: Volume in mL of tubing line between pump and dispense head for 1 line
    :type line_volume: int, float
    :param waste: Location of waste basin
    :type waste: class 'Well'

    """

    def __init__(
        self, index, name, pump_group, dispense_tip_offsets, line_volume, waste=None
    ):
        """
        Dispenser tool head for use with peristaltic pumps.

        Requires science_jubilee.tools.PeristalticPump.PeristalticPumps object to control pumps

        :param index: The tool index of the dispenser head on the machine
        :type indes: int
        :param name: The tool name
        :type name: str
        :param pump_group: science_jubilee.tools.PeristalticPump.PeristalticPumps object to assign to dispenser head
        :type pump_group: science_jubilee.tools.PeristalticPump.PeristalticPumps
        :param dispense_tip_offset: Offsets from tool reference point for every dispense tip on the tool
        :type dispense_tip_offset: list
        :param line_volume: Volume in mL of tubing line between pump and dispense head for 1 line
        :type line_volume: int, float
        :param waste: Location of waste basin
        :type waste: class 'Well'
        """
        super().__init__(
            index,
            name,
            pump_group=pump_group,
            dispense_tip_offsets=dispense_tip_offsets,
        )

        self.n_dispense_heads = len(self.dispense_tip_offsets)
        self.waste = waste
        self.line_volume = line_volume
        assert (
            self.pump_group.n_pumps == self.n_dispense_heads
        ), "Number of pumps must match number of dispenser tips"

    @classmethod
    def from_config(
        cls,
        index,
        pump_group,
        config_file: str,
        path: str = os.path.join(os.path.dirname(__file__), "configs"),
    ):
        """Initialize the pipette object from a config file

        :param index: The tool index of the dispenser_head on the machine
        :type index: int
        :param config_file: The name of the config file containign the dispenser tool parameters
        :type config_file: str
        :returns: A :class:`PumpDispenser` object
        :rtype: :class:`PumpDispenser`
        """
        config = os.path.join(path, config_file)
        with open(config) as f:
            kwargs = json.load(f)
        print(kwargs)
        kwargs["pump_group"] = pump_group
        kwargs["index"] = index

        return cls(**kwargs)

    def add_waste(self, location: Union[Well, Tuple, Location]):
        """
        Specify a waste collection container to dump extra liquid into. This should be large enough to catch liquid from all dispenser tips simultaneously.

        :param location: location of container
        :type location: class 'Well'
        """
        assert (
            isinstance(location, Well)
            or isinstance(location, Tuple)
            or isinstance(location, Location)
        ), "location must be a well, tuple, or location"
        self.waste = location

    def dispense(
        self,
        vol: Union[float, int, list],
        location: Union[Well, Tuple, Location],
        dispense_head_index=None,
    ):
        """
        Dispense volume vol from dispense head dispense_head_index into location

        :param vol: Volume to dispense, in mL. Specify a single number (float, int) and a dispense head index (int) to dispense from a single specific dispense head, the single volume with no dispense head to dispense this amount from all dispense tips, or a list of volumes to dispense unique amount from each dispense tip
        :type vol: float, int list
        :param location: Location to dispense
        :type location: class 'Well'
        :param dispense_head_index: index of dispense tip to dispense from, if passing single number for vol
        :type dispense_head_index: int
        """
        # get dispense volumes for each dispense head
        if isinstance(vol, list):
            assert len(vol) == self.n_dispense_heads
            dispense_volumes = vol
        elif isinstance(vol, float) or isinstance(vol, int):
            if dispense_head_index is None:
                # dispense same volume from every head
                dispense_volumes = [vol] * self.n_dispense_heads
            elif isinstance(dispense_head_index, int):
                dispense_volumes = [0] * self.n_dispense_heads
                dispense_volumes[dispense_head_index] = vol
            else:
                raise AssertionError("dispense_head_index must be an integer")
        else:
            raise AssertionError("vol must be a float, int, or list")

        # calculate XY location for each dispense head
        x, y, z = Labware._getxyz(location)

        if type(location) == Well:
            if z == location.z:
                z = z + 10
            else:
                pass
        else:
            pass

        # iterate over each dispense head, skip if 0. Move to location and dispense
        for i, vol in enumerate(dispense_volumes):
            if vol == 0:
                continue
            else:
                # get volume for this specific dispense head
                iter_vol = [0] * self.n_dispense_heads
                iter_vol[i] = vol

                tip_offsets = self.dispense_tip_offsets[i]
                x_tip = x + tip_offsets[0]
                y_tip = y + tip_offsets[1]

                self._machine.safe_z_movement()
                self._machine.move_to(x=x_tip, y=y_tip)
                self._machine.move_to(z=z)

                self.pump_group.pump(iter_vol)

    def prime_lines(
        self,
        volume: Union[int, float] = None,
        location: Union[Well, Tuple, Location] = None,
    ):
        """
        Fill all lines with liquid from all pumps simultaneously.

        :param volume: volume to prime with, if not specified in config file
        :type volue: int, float
        :param location: location of waste container, if not already specified
        :type location: class 'Well'
        """

        if volume is None:
            try:
                volume = self.line_volume
            except AttributeError:
                raise AssertionError(
                    "Line prime volume not specified in configuration. Please specify a volume"
                )

        if location is None and self.waste is None:
            raise AssertionError("Please specify a waste location")

        if location is None:
            location = self.waste

        # calculate XY location for each dispense head
        x, y, z = Labware._getxyz(location)

        if type(location) == Well:
            if z == location.z:
                z = z + 10
            else:
                pass
        else:
            pass

        self._machine.safe_z_movement()
        self._machine.move_to(x=x, y=y)
        self._machine.move_to(z=z)

        self.pump_group.pump(volume)

    def empty_lines(self, volume: Union[int, float] = None):
        """
        return line contents to stock wells.

        :param volume: line volume to return, if not set in instance already
        :type volume: int, float
        """
        if volume is None:
            try:
                volume = self.line_volume
            except AttributeError:
                raise AssertionError(
                    "Line prime volume not specified in configuration. Please specify a volume"
                )

        self.pump_group.pump(-volume)



import json
import logging
import os
from typing import Union

from science_jubilee.tools.Tool import Tool


class PeristalticPumps(Tool):
    """
    class representation of peristaltic pumps. Due to duet-level axis definition requirements, currently requires an accompanying PumpDispenser tool to control pumps.

    :param n_pumps: Number of peristaltic pumps
    :type n_pumps: int
    :param steps_per_mL: stepper motor steps to turn pump to pump 1mL of liquid. calibrate gravimetrically. Can pass a float or int to apply same value to each pump or list to apply unique value to each pump.
    :type steps_per_mL: float, int, list
    :param tool_axis: Motion axis tool is defined to on duet. Default E
    :type tool_axis: str
    :param name: name of tool
    :type name: str
    """

    def __init__(
        self,
        n_pumps: int,
        steps_per_ml: Union[float, int, list],
        tool_axis: str = "E",
        name: str = "dispenser_pumps",
    ):
        self.tool_axis = tool_axis
        self.n_pumps = n_pumps
        self.name = name

        if isinstance(steps_per_ml, float) or isinstance(steps_per_ml, int):
            self.steps_per_ml = [steps_per_ml] * n_pumps
        elif isinstance(steps_per_ml, list):
            assert (
                len(steps_per_ml) == n_pumps
            ), "steps per ml list length much equal number of pumps"
            self.steps_per_ml = steps_per_ml

    @classmethod
    def from_config(
        cls,
        config_file: str,
        path: str = os.path.join(os.path.dirname(__file__), "configs"),
    ):
        """Initialize a PeristalticPumps object from a config file

        :param config_file: The name of the config file containign the pump parameters
        :type config_file: str
        :returns: A :class:`PeristalticPumps` object
        :rtype: :class:`PeristalticPumps`
        """
        config = os.path.join(path, config_file)
        with open(config) as f:
            kwargs = json.load(f)
        print(kwargs)
        return cls(**kwargs)

    def post_load(self):
        """
        After tool load, set steps per mm on machine
        """
        # set steps per mL on machine
        steps_per_ml = self.steps_per_ml

        gcode = f"M92 {self.tool_axis}" + ":".join(
            [str(stpml) for stpml in steps_per_ml]
        )
        print(gcode)
        self._machine.gcode(gcode)

        return

    def pump(self, volume: Union[int, float, list]):
        """transfer volume using pump

        :param volume: volume to pump. Specify a number (int/float) to pump same amount from all pumps simultaneously, or a list of numbers to pump unique amount from each pump.
        :type volume: int, float, list
        """
        # dispense given volume
        if isinstance(volume, list):
            assert isinstance(volume[0], float) or isinstance(
                volume[0], int
            ), "Volume list must be floats or ints"
            pass

        elif isinstance(volume, float) or isinstance(volume, int):
            volume = [volume] * self.n_pumps

        else:
            raise TypeError("Volume must be an int, float, or list of ints or floats")

        stringvol = ":".join([str(v) for v in volume])

        # sticking with a direct gcode to send here makes sense: no 3-motor on axis support in movement code, and no risk of crashing anything here
        self._machine.gcode(f"G1 {self.tool_axis}{stringvol}")



from pathlib import Path


class PlateStateError(Exception):
    """Raise this error if the plate is in the wrong state to perform such a command."""

    pass


class PlateConfigurationError(Exception):
    """Raise this error if there is something wrong with how the plate is configured"""

    pass


class Plate:
    def __init__(self, machine, name):
        self._machine = machine
        self._name = name

    def get_root_dir(self):
        return Path(__file__).parent.parent.parent



import json
import os
import random
import warnings
from typing import Tuple

import numpy as np

from science_jubilee.labware.Labware import Well
from science_jubilee.tools.Tool import Tool, requires_active_tool


class Loop(Tool):
    """A class representation of an inoculation loop.

    :param Tool: The base tool class
    :type Tool: :class:`Tool`
    """

    def __init__(self, index, name):
        """Constructor method"""
        super().__init__(index, name)

    @requires_active_tool
    def transfer(
        self,
        source: Well = None,
        destination: Well = None,
        s: int = 2000,
        sweep_x: float = 5,
        sweep_y: float = 5,
        sweep_z: float = 10,
        sweep_speed: float = 100,
        up_speed: float = 800,
        randomize_pickup: bool = False,
    ):
        """Dip into a number of (source, destination) well pairs. Used for transferring items from source to destination.

        :param source: Source well, defaults to None
        :type source: :class:`Well`, optional
        :param destination: Destination well, defaults to None
        :type destination: :class:`Well`, optional
        :param s: Movement speed in mm/min, defaults to 2000
        :type s: int, optional
        :param sweep_x: Distance (in mm) to sweep in the x direction to pick up material, defaults to 5
        :type sweep_x: float, optional
        :param sweep_y: Distance (in mm) to sweep in the y direction to pick up material, defaults to 5
        :type sweep_y: float, optional
        :param sweep_z: Distance (in mm) to move in the z direction to pick up material, defaults to 10
        :type sweep_z: float, optional
        :param sweep_speed: Speed (in mm/min) at which to sweep to pick up material, defaults to 100
        :type sweep_speed: float, optional
        :param up_speed: Speed to move out of well (in mm/min), defaults to 800
        :type up_speed: float, optional
        :param randomize_pickup: Move to a random position in the source well to pick up material, defaults to False
        :type randomize_pickup: bool, optional
        """
        if type(source) != list:
            source = [source]
        if type(destination) != list:
            destination = [destination]

        # Assemble tuples of (source, destination)
        num_source_wells = len(source)
        num_destination_wells = len(destination)
        if num_source_wells == num_destination_wells:  # n to n transfers
            pass
        elif (
            num_source_wells == 1 and num_destination_wells > 1
        ):  # one to many transfers
            source = list(np.repeat(source, num_destination_wells))
        elif (
            num_source_wells > 1 and num_destination_wells == 1
        ):  # many to one transfers
            destination = list(np.repeat(destination, num_source_wells))
        elif num_source_wells > 1 and num_destination_wells > 1:  # uneven transfers
            # for uneven transfers, find least common multiple to pair off wells
            # raise a warning, as this might be a mistake
            # this mimics OT-2 behavior
            least_common_multiple = np.lcm(num_source_wells, num_destination_wells)
            source_repeat = least_common_multiple / num_source_wells
            destination_repeat = least_common_multiple / num_destination_wells
            source = list(np.repeat(source, source_repeat))
            destination = list(np.repeat(destination, destination_repeat))
            warnings.warn("Warning: Uneven source & destination wells specified.")

        source_destination_pairs = list(zip(source, destination))
        for source_well, destination_well in source_destination_pairs:
            xs, ys, zs = self._get_xyz(well=source_well)
            if (
                randomize_pickup
            ):  # to make sure we don't try to pickup from an empty region
                r = 20
                rx = random.randint(-r, r)
                ry = random.randint(-r, r)
                xs += rx
                ys += ry
            xd, yd, zd = self._get_xyz(well=destination_well)

            self._machine.safe_z_movement()
            self._machine.move_to(x=xs, y=ys)
            self._machine.move_to(z=zs + 5)
            # slowly sweep in the reservoir to pick up duckweed
            # can tune these default values
            self._machine.move(dx=sweep_x, s=sweep_speed)
            self._machine.move(dy=sweep_y, s=sweep_speed)
            self._machine.move(dz=sweep_z, s=up_speed)
            self.current_well = source_well
            # self._aspirate(vol, s=s)

            self._machine.safe_z_movement()
            self._machine.move_to(x=xd, y=yd)
            self._machine.move_to(z=zd + 5)
            # sweep again to drop off duckweed
            # make smaller movements and move opposite direction
            self._machine.move(dx=sweep_x / 2, s=sweep_speed)
            self._machine.move(dy=-sweep_y, s=sweep_speed)
            self._machine.dwell(250)  # give the duckweed time to release
            self.current_well = destination_well
            # self._dispense(vol, s=s)

    @staticmethod
    def _get_xyz(well: Well = None, location: Tuple[float] = None):
        """Get the (x,y,z) position of a well.

        :param well: The well to fetch position of, defaults to None
        :type well: :class:`Well`, optional
        :param location: Directly specify an (x,y,z) location, defaults to None
        :type location: Tuple[float], optional
        :raises ValueError: Must specify either a well or a location
        :return: The well location
        :rtype: Tuple[float, float, float]
        """
        if well is not None and location is not None:
            raise ValueError("Specify only one of Well or x,y,z location")
        elif well is not None:
            x, y, z = well.x, well.y, well.z
        else:
            x, y, z = location
        return x, y, z

    @staticmethod
    def _get_top_bottom(well: Well = None):
        """Get the top and bottom heights of a well.

        :param well: The well to fetch position of, defaults to None
        :type well: Well, optional
        :return: The z-height of the top and bottom of the well
        :rtype: Tuple[float, float]
        """
        top = well.top
        bottom = well.bottom
        return top, bottom



import json
import math
import os
import warnings
from typing import Tuple, Union

import numpy as np

from science_jubilee.labware.Labware import Labware, Location, Well
from science_jubilee.tools.Tool import (
    Tool,
    ToolConfigurationError,
    ToolStateError,
    requires_active_tool,
)


class SyringeExtruder(Tool):
    """A class representation of a syringe for extrusion 3D printing.

    :param Tool: The base tool class
    :type Tool: :class:`Tool`
    """

    def __init__(self, index, name, config):
        """Constructor method"""
        super().__init__(index, name)

        self.min_range = 0
        self.max_range = None
        self.mm_to_ml = None
        self.e_drive = "E1"

        self.nozzle_diameter = 0.898  # default 18 gauge tip
        self.syringe_diameter = 14

        self.load_config(config)

    def load_config(self, config):
        """Loads the confirguration file for the syringe tool

        :param config: Name of the config file for your syringe. Expects the file to be in /tools/configs
        :type config: str
        """

        config_directory = os.path.join(os.path.dirname(__file__), "configs")
        config_path = os.path.join(config_directory, f"{config}.json")
        if not os.path.isfile(config_path):
            raise ToolConfigurationError(
                f"Error: Config file {config_path} does not exist!"
            )

        with open(config_path, "r") as f:
            config = json.load(f)
        self.min_range = config["min_range"]
        self.max_range = config["max_range"]
        self.mm_to_ml = config["mm_to_ml"]

        # Check that all information was provided
        if None in vars(self):
            raise ToolConfigurationError(
                "Error: Not enough information provided in configuration file."
            )

    def post_load(self):
        """Query the object model after loading the tool to find the extruder number of this syringe."""

        # To read the position of an extruder, we need to know which extruder number to look at
        # Query the object model to find this
        tool_info = json.loads(self._machine.gcode('M409 K"tools[]"'))["result"]
        for tool in tool_info:
            if tool is None:
                continue
            if tool["number"] == self.index:
                self.e_drive = (
                    f"E{tool['extruders'][0]}"  # Syringe tool has only 1 extruder
                )
            else:
                continue

    def check_bounds(self, pos):
        """Disallow commands outside of the syringe's configured range

        :param pos: The E position to check
        :type pos: float
        """
        if pos > self.max_range or pos < self.min_range:
            raise ToolStateError(f"Error: {pos} is out of bounds for the syringe!")

    def make_e(self, x, y, z):
        pos = self._machine.get_position()
        start = [float(pos["X"]), float(pos["Y"]), float(pos["Z"])]
        if x is None:
            x = start[0]
        if y is None:
            y = start[1]
        if z is None:
            z = start[2]
        end = [x, y, z]
        d = self.dist(start, end)

        return 2 * d * (self.nozzle_diameter / self.syringe_diameter) ** 2

    def dist(self, start, end):
        dist = math.sqrt(
            (end[0] - start[0]) ** 2
            + (end[1] - start[1]) ** 2
            + (end[2] - start[2]) ** 2
        )
        return dist

    def wipe_nozzle(self, x=285, y=250, z=0.2):
        self._machine.move_to(x=x, y=y)
        self._machine.move_to(z=z)
        self.move_extrude(x=x, y=y - 50, z=z, multiplier=3)

    def wipe_tower(self, x=285, y=250, z=0.2):
        self._machine.move_to(x=x, y=y, s=3600)
        self._machine.move_to(z=z)
        self.move_extrude(x=x, y=y - 20, z=z, s=200)
        self.move_extrude(x=x - 20, y=y - 20, z=z, s=200)
        self.move_extrude(x=x - 20, y=y, z=z, s=200)
        self.move_extrude(x=x, y=y, z=z, s=200)
        self._machine.move(dx=3)

    @requires_active_tool
    def move_extrude(self, x=None, y=None, z=None, s=180, multiplier=1, e=None):
        if not e:
            e = self.make_e(x, y, z)
            e *= multiplier

        x = "{0:.2f}".format(x) if x is not None else None
        y = "{0:.2f}".format(y) if y is not None else None
        z = "{0:.2f}".format(z) if z is not None else None
        e = "{0:.2f}".format(e) if e is not None else None
        s = "{0:.2f}".format(s)

        # initialize coordinates commands
        x_cmd = y_cmd = z_cmd = e_cmd = f_cmd = ""

        if x is not None:
            x_cmd = f"X{x}"
        if y is not None:
            y_cmd = f"Y{y}"
        if z is not None:
            z_cmd = f"Z{z}"
        if e is not None:
            e_cmd = f"E{e}"
        if s is not None:
            f_cmd = f"F{s}"

        cmd = f"G0 {z_cmd} {x_cmd} {y_cmd} {e_cmd} {f_cmd}"
        self._machine.gcode(cmd)

    @requires_active_tool
    def _aspirate(self, vol: float, s: int = 2000):
        """Aspirate a certain volume in milliliters. Used only to move the syringe; to aspirate from a particular well, see aspirate()

        :param vol: Volume to aspirate, in milliliters
        :type vol: float
        :param s: Speed at which to aspirate in mm/min, defaults to 2000
        :type s: int, optional
        """
        de = vol * -1 * self.mm_to_ml
        pos = self._machine.get_position()
        end_pos = float(pos[self.e_drive]) + de
        self.check_bounds(end_pos)
        self._machine.move(de=de, wait=True)

    @requires_active_tool
    def _dispense(self, vol, s: int = 2000):
        """Dispense a certain volume in milliliters. Used only to move the syringe; to dispense into a particular well, see dispense()

        :param vol: Volume to dispense, in milliliters
        :type vol: float
        :param s: Speed at which to dispense in mm/min, defaults to 2000
        :type s: int, optional
        """
        de = vol * self.mm_to_ml
        pos = self._machine.get_position()
        end_pos = float(pos[self.e_drive]) + de
        self.check_bounds(end_pos)
        self._machine.move(de=de, wait=True)

    @requires_active_tool
    def aspirate(
        self, vol: float, location: Union[Well, Tuple, Location], s: int = 2000
    ):
        """Aspirate a certain volume from a given well.

        :param vol: Volume to aspirate, in milliliters
        :type vol: float
        :param location: The location (e.g. a `Well` object) from where to aspirate the liquid from.
        :type location: Union[Well, Tuple, Location]
        :param s: Speed at which to aspirate in mm/min, defaults to 2000
        :type s: int, optional
        """
        x, y, z = Labware._getxyz(location)

        # self._machine.safe_z_movement()
        self._machine.move_to(x=x, y=y)
        self._machine.move_to(z=z)
        self._aspirate(vol, s=s)

    @requires_active_tool
    def dispense(
        self, vol: float, location: Union[Well, Tuple, Location], s: int = 2000
    ):
        """Dispense a certain volume into a given well.

        :param vol:  Volume to dispense, in milliliters
        :type vol: float
        :param location: The location to dispense the liquid into.
        :type location: Union[Well, Tuple, Location]
        :param s: Speed at which to dispense in mm/min, defaults to 2000
        :type s: int, optional

        """
        x, y, z = Labware._getxyz(location)

        # self._machine.safe_z_movement()
        self._machine.move_to(x=x, y=y)
        self._machine.move_to(z=z)
        self._dispense(vol, s=s)



import json
import logging
import os
from datetime import date
from typing import Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

# this is the Ocean Optics SDK, which is (very unfortunately) not open-source
try:
    from oceandirect.OceanDirectAPI import OceanDirectAPI
except ImportError:
    raise ImportError(
        "The Ocean Optics SDK is not installed. Please install it from the Ocean Insight website."
    )

from science_jubilee.labware.Labware import Labware, Location, Well
from science_jubilee.tools.Tool import Tool, requires_active_tool


class SpectroscopyTool(Tool, OceanDirectAPI):
    """A class representation for an Ocean Optics Spectrometer, Light source, and Probe

    For that we wil be unfortunately using their own SDK package, which is NOT
    open-source."""

    def __init__(self, index, name):
        self.connection = OceanDirectAPI()
        self.spectrometer = self._open_device()
        self.name = name
        self.index = index
        self.current_well = None
        self._dark_spectrum = False
        self.reference_spectrum = None

    @property
    def _api_version(self):
        """The version of the Ocean Insight SDK"""
        (major, minor, point) = self.connection.get_api_version_numbers()
        API_vers = "%d.%d.%d" % (major, minor, point)
        return API_vers

    @property
    def _usb_device(self):
        devices_found = self.connection.find_usb_devices()
        return devices_found

    @property
    def _device_id(self):
        """The ID of the spectrometer"""
        usb = self._usb_device
        ids = self.connection.get_device_ids()
        return ids

    def _open_device(self):
        """Connects to the spectrometer device"""
        devices = []
        idx = self._device_id
        for i in idx:
            dev = self.connection.open_device(i)
            devices.append(dev)

        tot_dev = len(devices)
        if tot_dev == 1:
            self.spectrometer = devices[0]
            return devices[0]
        else:
            print(
                "Careful, more than one (%d) Ocean Insight device is currently"
                + "connected to this device." % tot_dev
            )
            return devices

    @property
    def _model(self):
        """The spectrometer model"""
        model = self.spectrometer.get_model()
        return model

    @property
    def _serial_number(self):
        """The spectrometer serial number"""
        sn = self.spectrometer.get_serial_number()
        return sn

    @property
    def _integration_time(self):
        it = self.spectrometer.get_integration_time()
        return it

    def integration_time(self, t: float, units: str = "ms"):
        """Sets the integration time for the spectrometer.

        :param t: integration time
        :type t: float
        :param units: units of integration time, defaults to 'ms'
        :type units: str, optional
        """
        self._integration_time_units = units

        # check provided integration time units are supported by the code
        supported_units = ["us", "ms", "s"]
        assert (
            units in supported_units
        ), "Indicate integration time in one of the supported units: %s, %s, %s" % (
            supported_units[0],
            supported_units[1],
            supported_units[2],
        )
        if units == "ms":
            time = t * 10e3
        elif units == "s":
            time = t * 10e6
        else:
            time = t

        tint_min = self.spectrometer.get_minimum_integration_time()
        tint_max = self.spectrometer.get_maximum_integration_time()
        assert (
            tint_min <= time <= tint_max
        ), "Indicate integration time between %d and %d ms" % (tint_min, tint_max)

        time = int(time)

        self.spectrometer.set_integration_time(time)
        print("The integration time was set to: %d %s" % (time, units))

        return

    @property
    def wavelengths(self):
        """Return list containing the wavelengths of the spectrometer.

        :return: A list of wavelengths
        :rtype: list
        """
        wlt = self.spectrometer.wavelengths
        return wlt

    def scans_to_average(self, n: int):
        """Set the number of scans to average together to compose a single spectrum.

        :param n: number of scans to average
        :type n:int

        """
        self.spectrometer.set_scans_to_average(n)

        return print("The number of scans to average was set to: %d" % n)

    def boxcar_width(self, w: int = None):
        """Unit used to smooth the spectrum.

        The boxcar width is the number of adjacent pixels that are averaged together to smooth the spectrum.
        Pixel average = 2* boxcar_width + 1

        :param w: boxcar width
        :type w: int
        """
        if w is not None:
            self.spectrometer.set_boxcar_width(w)
        else:
            w = self.spectrometer.get_boxcar_width()

        print("The boxcar width was set to: %d" % w)
        return

    def lamp_shutter(self, open=True):
        """Opens or closes light source shutter, if feature is available on light source.)

        :param open: True opens the shutter/False closes it, defaults to False
        :type state: bool, optional
        """
        # need to add a try statement as not all light-sources have a shutter
        if open == True:
            state = "Open"
        else:
            state = "Close"

        self.spectrometer.Advanced.set_enable_lamp(open)

        if self.spectrometer.Advanced.get_enable_lamp() == open:
            print(f"Light shutter was set to {state}")
        else:
            print("No such feature on current device")
        # except:
        #     # logging.warning('No light source shutter available')
        #     print('No such feature on current device')
        return

    def _take_spectrum(self, open=True):
        """Collect a spectrum with the specified experimental parameters"""

        self.lamp_shutter(open=open)

        spectrum = self.spectrometer.get_formatted_spectrum()

        return spectrum

    def dark_spectrum(
        self,
        int_time: float,
        scan_num: int,
        boxcar_w: int = None,
        save: bool = False,
        path: str = None,
        filename: str = None,
    ):
        """Collect a dark spectrum with the specified experimental parameters.

        A dark spectrum is collected when no light is incident on the detector.
        This shuold later be used to correct the actual spectrum for dark counts (i.e. thermal noise).
        Note: these shuold be the same as the ones used for the actual spectrum.

        :param int_time: integration time
        """

        self.integration_time(int_time)
        self.scans_to_average(scan_num)
        self.boxcar_width(boxcar_w)

        dark = self._take_spectrum(open=False)

        self._dark_spectrum = True
        self._dark = dark

        if save is True:
            self.save_to_file(dark, dark=True, path=path, filename=filename)

        return dark

    def _collect_raw_spectrum(
        self, int_time: float, scan_num: int, boxcar_w: int, int_time_units: str = "ms"
    ):
        """Collect a spectrum with the specified experimental parameters

        :param int_time: integration time
        :type int_time: float
        :param scan_num: number of scans to average
        :type scan_num: int
        :param boxcar_w: boxcar width for smoothing the spectrum, defaults to None
        :type boxcar_w: int, optional
        :param int_time_units: units of integration time, defaults to 'ms'
        :type int_time_units: str, optional
        :return: spectrum
        :rtype: numpy array
        """

        self.integration_time(int_time, units=int_time_units)
        self.scans_to_average(scan_num)
        self.boxcar_width(boxcar_w)

        if self._dark_spectrum is False:
            self.dark_spectrum(int_time, scan_num, boxcar_w)
        else:
            pass

        spectrum = self._take_spectrum(open=True)

        return spectrum

    @requires_active_tool
    def collect_spectrum(
        self,
        location: Union[Well, Tuple, Location],
        int_time,
        scan_num,
        boxcar_w,
        int_time_units="ms",
        save: bool = False,
        filename: str = None,
        path: str = None,
    ):
        """Collect spectrum at the specified location on the labware"""

        x, y, z = Labware._getxyz(location)

        if type(location) == Well:
            self.current_well = location
        elif type(location) == Location:
            self.current_well = location._labware
        else:
            pass

        self._machine.safe_z_movement()
        self._machine.move_to(x=x, y=y)
        self._machine.move_to(z=z)
        intensities = self._collect_raw_spectrum(
            int_time, scan_num, boxcar_w, int_time_units=int_time_units
        )

        if save is True:
            self.save_to_file(intensities, dark=False, path=path, filename=filename)

        return intensities

    def save_to_file(
        self, data, dark: bool = False, path: str = None, filename: str = None
    ):
        """Save the spectrum to a file"""

        metadata = {
            "Date": date.today().strftime("%Y%m%d"),
            "Tool": self.name,
            "Index": self.index,
            "Model": self._model,
            "Serial Number": self._serial_number,
            "API Version": self._api_version,
            "USB Device": self._usb_device,
            "Device ID": self._device_id,
            "Integration Time": self._integration_time,
        }

        raw_data = {"wavelength": self.wavelengths, "intensity": data}

        if filename is None:
            if dark is True:
                filename = f"dark_spectrum_{metadata['Date']}.txt"
            else:
                filename = f"slot{self.current_well.slot}_{self.current_well.name}_spectrum_{metadata['Date']}.txt"
        else:
            pass

        if path is None:
            path = "./"
        else:
            pass
        # add well location information to metadata as well if not the 'dark' spectrum
        if dark is not True:
            metadata["Well"] = self.current_well.name
            metadata["Slot"] = self.current_well.slot
            data_handling = MeasurementManager(metadata, raw_data)
            data_handling.generate_file(filename, path)

        print(f"Spectrum saved to {path}/{filename}")
        return

    def read_from_file(self, filepath, scale=True, reference_spectrum=None):

        metadata, measurements = MeasurementManager.read_file(filepath)

        if scale == True:
            if reference_spectrum == None and self.reference_spectrum == None:
                print("Reference spectrum must be provided for scaling.")
                return
            elif reference_spectrum == None and self.reference_spectrum != None:
                reference_spectrum = self.reference_spectrum
            else:
                pass
            measurements["Intensity"] = self.scale_intensity(
                measurements["Intensity"], reference_spectrum
            )
        return metadata, measurements

    @staticmethod
    def scale_intensity(data: list, reference_spectrum: list):

        assert len(data) == len(
            reference_spectrum
        ), "Data and reference spectrum must be of same size"
        scaled_intensities = [d / r for d, r in zip(data, reference_spectrum)]

        return scaled_intensities


class MeasurementManager:
    """A class to manage the measurements taken by the SpectroscopyTool"""

    def __init__(self, metadata, raw_data):
        self.raw_data = raw_data
        self.wavelengths = raw_data["wavelength"]
        self.intensities = raw_data["intensity"]
        self.metadata = metadata

    def pretty_print(self):
        print("---- METADATA ----")
        for key, value in self.metadata.items():
            print(f"{key}: {value}")
        print("\n---- RAW DATA ----")
        print(self.raw_data)

    def generate_file(self, filename, path):
        if not filename.endswith(".txt"):
            filename += ".txt"
        filepath = os.path.join(path, filename)

        with open(filepath, "w") as file:
            # Write metadata section
            file.write("---- METADATA ----\n")
            for key, value in self.metadata.items():
                file.write(f"{key}: {value}\n")

            # Write raw data section
            file.write("---- RAW DATA ----\n")
            for wavelength, intensity in zip(self.wavelengths, self.intensities):
                file.write(f"{wavelength}\t{intensity}\n")

    def read_file(filepath):
        metadata = {}
        measurements = {"Wavelength": [], "Intensity": []}

        with open(filepath, "r") as file:
            section = None
            for line in file:
                line = line.strip()
                if line.startswith("---- METADATA ----"):
                    section = "metadata"
                elif line.startswith("---- RAW DATA ----"):
                    section = "raw_data"
                elif section == "metadata":
                    key, value = line.split(": ")
                    metadata[key] = value
                elif section == "raw_data":
                    wavelength, intensity = line.split("\t")
                    measurements["Wavelength"].append(float(wavelength))
                    measurements["Intensity"].append(float(intensity))

        return metadata, measurements



import json
import os
from dataclasses import dataclass
from typing import Dict, Tuple

from science_jubilee.labware.Labware import Labware


@dataclass
class Slot:
    """Class that defines a slot on the Jubilee deck.
    Each slot has a unique index, offset, and can have a labware associated with it.

    :param slot_index: The unique index of the slot, usually a number between 0 and 5.
    :type slot_index: int
    :param offset: The (x,y) offset of the slot from the origin of the deck/machine.
    :type offset: Tuple[float]
    :param has_labware: A boolean that indicates whether a labware has been loaded into the slot.
    :type has_labware: bool
    :param labware: The :class:`Labware` object that has been loaded into the slot.
    :type labware: :class:`Labware`
    """

    slot_index: int
    offset: Tuple[float]
    has_labware: bool
    labware: str


@dataclass
class SlotSet:
    """Class that defines a set of slots on the Jubilee deck.

    :param slots: A dictionary of :class:`Slot` objects, where the key is the slot's index.
    :type slots: Dict[str, :class:`Slot`]
    :param bedType: The type of bed that the slots are arranged in. Usually 'fixed' or 'removable'.
    :type bedType: str

    """

    slots: Dict[str, Slot]

    def __repr__(self):
        return str(self.bedType)

    def __getitem__(self, id_):
        """Allows the user to select a :class:`Slot` object by its index or key.

        :param id_: The index or key representing the slot.
        :type id_: str or int
        :return: The :class:`Slot` object associated with the index or key.
        :rtype: :class:`Slot`
        """
        try:
            return self.slots[id_]
        except KeyError:
            return list(self.slots.values())[id_]


class Deck(SlotSet):
    """Class that defines the Jubilee deck.

    The deck is a set of slots that can have labware loaded into them.
    The deck is defined by a configuration file that specifies the number of slots, their offsets, and the type of bed they are arranged in.

    :param deck_filename: The name of the deck configuration file.
    :type deck_filename: str
    :param path: The path to the deck configuration `.json` files for the labware,
            defaults to the 'deck_definition/' in the science_jubilee/decks directory.
    :type path: str, optional
    """

    def __init__(
        self,
        deck_filename,
        path: str = os.path.join(os.path.dirname(__file__), "deck_definition"),
    ):
        """Initializes the :class:`Deck` object by loading its configuration file and creating a dictionary of :class:`Slot` objects.

        :param deck_filename: The name of the deck configuration file.
        :type deck_filename: str
        :param path: The path to the deck configuration `.json` files for the labware,
                defaults to the 'deck_definition/' in the science_jubilee/decks directory.
        :type path: str, optional
        """
        # load in the deck configuration file
        if deck_filename[-4:] != "json":
            deck_filename = deck_filename + ".json"

        config_path = os.path.join(path, f"{deck_filename}")

        with open(config_path, "r") as f:
            deck_config = json.load(f)

        self.deck_config = deck_config
        self.slots_data = self.deck_config.get("slots", {})
        self.slots = self._get_slots()
        self._safe_z = 10

    def _get_slots(self):
        """Function that creates a dictionary of :class:`Slot` objects from the deck configuration file.

        :return: A dictionary of :class:`Slot` objects, where the key is the slot's index.
        :rtype: Dict[str, :class:`Slot`]
        """
        slots = {}
        for s, sv in self.slots_data.items():
            if type(sv) == list:
                sv = tuple(sv)
            else:
                pass
            slots[s] = Slot(slot_index=s, **self.slots_data[s])
        return slots

    @property
    def bedType(self):
        """Function that returns the type of bed loaded onto Jubilee.

        :return: The name/type of deck loaded onto Jubilee, e.g., Lab Automation Deck, Heated Deck, etc.
        :rtype: str
        """

        return self.deck_config.get("bedType", "")

    @property
    def totalslots(self):
        """Function that returns the total number of slots on the deck.

        :return: The total number of slots on the deck.
        :rtype: int
        """

        deckslots = self.deck_config.get("deckSlots", {})
        return deckslots["total"]

    @property
    def slotType(self):
        """Function that returns the type of slot arrangement the deck might have.

        :return: The slot arrangement type. This is inidcated in the configuration file. Standard is "SLAS".
        :rtype: str
        """

        deckslots = self.deck_config.get("deckSlots", {})
        return deckslots["type"]

    @property
    def offsetFrom(self):
        """Function that returns which corner or the slot to apply to a labware loaded on it.

        :return: The corner of the slot to apply to the labware. This is inidcated in the configuration file.
        :rtype: str
        """
        return self.deck_config.get("offsetFrom", {})

    @property
    def deck_material(self):
        """Function that returns the material that the deck and possible mask are made of.

        :return: The material that the deck is made of, as well as any mask that is applied to it.
        :rtype: Dict[str, str]
        """
        return self.deck_config.get("material", {})

    @property
    def safe_z(self):
        """Function that returns the movement clearance height of the deck.

        :return: The height at which the pipette can freely move without colliding with
            labware on the deck.
        :rtype: float
        """
        return self._safe_z

    @safe_z.setter
    def safe_z(self, val):
        """Function that updates the movement clearance height
        every time a new labware is loaded onto the deck

        :param val: The new safe z height.
        :type val: float
        """
        if self._safe_z is None:
            self._safe_z = val
        elif self._safe_z <= val:
            self._safe_z = val
        else:
            pass

    def load_labware(
        self,
        labware_filename: str,
        slot: int,
        path=os.path.join(
            os.path.dirname(__file__), "..", "labware", "labware_definition"
        ),
        order: str = "rows",
    ):
        """Function that loads a labware and associates it with a specific slot on the deck.
         The slot offset is also applied to the labware asocaite with it.

        :param labware_filename: The name of the labware configuration file.
        :type labware_filename: str
        :param slot: The index of the slot to load the labware into.
        :type slot: int
        :param path: The path to the labware configuration `.json` files for the labware,
                defaults to the 'labware_definition/' in the science_jubilee/labware directory.
        :type path: str, optional
        :param order: The order in which the labware is arranged on the deck.
                Can be 'rows' or 'columns', defaults to 'rows'.
        :type order: str, optional
        :return: The :class:`Labware` object that has been loaded into the slot.
        :rtype: :class:`Labware`"""

        labware = Labware(labware_filename, order=order)
        labware.add_slot(slot)
        offset = self.slots[str(slot)].offset

        labware.offset = offset

        self.slots[str(slot)].has_labware = True
        self.slots[str(slot)].labware = labware
        self.safe_z = labware.dimensions["zDimension"]
        return labware
